# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

import numpy as np
cimport numpy as np

from libc.math cimport INFINITY
from libc.math cimport NAN

from libc.stdlib cimport malloc
from libc.stdlib cimport free

from libc.string cimport memcpy
from libc.string cimport memset

from ..distance._distance cimport TSDatabase
from ..distance._distance cimport DistanceMeasure

from ..distance._distance cimport TSView
from ..distance._distance cimport TSCopy

from ..distance._distance cimport ts_database_new

from ..distance._distance cimport ts_view_init
from ..distance._distance cimport ts_view_free
from ..distance._distance cimport ts_copy_free
from ..distance._distance cimport new_distance_measure

from ._impurity cimport entropy

from .._utils cimport label_distribution
from .._utils cimport argsort
from .._utils cimport rand_int
from .._utils cimport RAND_R_MAX
from .._utils cimport RollingVariance
from .._utils cimport safe_realloc
from .._utils cimport rand_uniform


cpdef Tree _make_tree(
    DistanceMeasure distance_measure,
    Py_ssize_t n_labels,
    list shapelets,
    np.ndarray threshold,
    np.ndarray value,
    np.ndarray left,
    np.ndarray right,
    np.ndarray impurity,
    np.ndarray n_node_samples,
    np.ndarray n_weighted_node_samples
):
    cdef Tree tree = Tree(distance_measure, n_labels, capacity=len(shapelets) + 1)
    cdef Py_ssize_t node_count = len(shapelets)
    cdef Py_ssize_t i
    cdef Py_ssize_t dim
    cdef np.ndarray arr
    cdef TSCopy *shapelet
    cdef np.ndarray value_reshape = value.reshape(-1)

    tree._node_count = node_count
    for i in range(node_count):
        if shapelets[i] is not None:
            dim, arr = shapelets[i]
            shapelet = <TSCopy*> malloc(sizeof(TSCopy))
            distance_measure.init_ts_copy_from_ndarray(shapelet, arr, dim)
            tree._shapelets[i] = shapelet
        else:
            tree._shapelets[i] = NULL
        tree._thresholds[i] = threshold[i]
        tree._left[i] = left[i]
        tree._right[i] = right[i]
        tree._impurity[i] = impurity[i]
        tree._n_node_samples[i] = n_node_samples[i]
        tree._n_weighted_node_samples[i] = n_weighted_node_samples[i]

    for i in range(node_count * n_labels):
        tree._values[i] = value_reshape[i]
    return tree


cdef class Tree:
    def __cinit__(
        self,
        DistanceMeasure distance_measure,
        Py_ssize_t n_labels,
        Py_ssize_t capacity=10
    ):
        self.distance_measure = distance_measure
        self._node_count = 0
        self._capacity = capacity
        self._n_labels = n_labels
        self._shapelets = <TSCopy**> malloc(self._capacity * sizeof(TSCopy*))
        self._thresholds = <double*> malloc(self._capacity * sizeof(double))
        self._values = <double*> malloc(self._capacity * self._n_labels * sizeof(double))
        self._left = <Py_ssize_t*> malloc(self._capacity * sizeof(Py_ssize_t))
        self._right = <Py_ssize_t*> malloc(self._capacity * sizeof(Py_ssize_t))
        self._impurity = <double*> malloc(self._capacity * sizeof(double))
        self._n_node_samples = <Py_ssize_t*> malloc(self._capacity * sizeof(Py_ssize_t))
        self._n_weighted_node_samples = <double*> malloc(self._capacity * sizeof(double))

    def __dealloc__(self):
        cdef Py_ssize_t i
        if self._shapelets != NULL:
            for i in range(self._node_count):
                if self._shapelets[i] != NULL:
                    free(self._shapelets[i])
                    ts_copy_free(self._shapelets[i])
            free(self._shapelets)

        if self._thresholds != NULL:
            free(self._thresholds)

        if self._values != NULL:
            free(self._values)

        if self._left != NULL:
            free(self._left)

        if self._right != NULL:
            free(self._right)

        if self._impurity != NULL:
            free(self._impurity)

        if self._n_node_samples != NULL:
            free(self._n_node_samples)

        if self._n_weighted_node_samples != NULL:
            free(self._n_weighted_node_samples)

    def __reduce__(self):
        return _make_tree, (
            self.distance_measure,
            self._n_labels,
            self.shapelet,
            self.threshold,
            self.value,
            self.left,
            self.right,
            self.impurity,
            self.n_node_samples,
            self.n_weighted_node_samples,
        )


    @property
    def max_depth(self):
        return self._max_depth

    @property
    def value(self):
        cdef np.ndarray arr = np.empty(self._node_count * self._n_labels, dtype=np.float64)
        cdef Py_ssize_t i
        for i in range(self._n_labels * self._node_count):
            arr[i] = self._values[i]
        return arr.reshape(self._node_count, self._n_labels)

    @property
    def shapelet(self):
        cdef TSCopy *shapelet
        cdef np.ndarray temp
        cdef Py_ssize_t i, j
        cdef list ret = []
        for i in range(self._node_count):
            shapelet = self._shapelets[i]
            if shapelet != NULL:
                temp = np.empty(shapelet[0].length, dtype=np.float64)
                for j in range(shapelet[0].length):
                    temp[j] = shapelet[0].data[j]
                ret.append((shapelet[0].dim, temp))
            else:
                ret.append(None)
        return ret

    @property
    def n_node_samples(self):
        cdef np.ndarray arr = np.zeros(self._node_count, dtype=np.intp)
        cdef Py_ssize_t i
        for i in range(self._node_count):
            arr[i] = self._n_node_samples[i]
        return arr

    @property
    def n_weighted_node_samples(self):
        cdef np.ndarray arr = np.zeros(self._node_count, dtype=np.float64)
        cdef Py_ssize_t i
        for i in range(self._node_count):
            arr[i] = self._n_weighted_node_samples[i]
        return arr

    @property
    def left(self):
        cdef np.ndarray arr = np.empty(self._node_count, dtype=np.intp)
        cdef Py_ssize_t i
        for i in range(self._node_count):
            arr[i] = self._left[i]
        return arr

    @property
    def right(self):
        cdef np.ndarray arr = np.empty(self._node_count, dtype=np.intp)
        cdef Py_ssize_t i
        for i in range(self._node_count):
            arr[i] = self._right[i]
        return arr

    @property
    def threshold(self):
        cdef np.ndarray arr = np.empty(self._node_count, dtype=np.float64)
        cdef Py_ssize_t i
        for i in range(self._node_count):
            arr[i] = self._thresholds[i]
        return arr

    @property
    def impurity(self):
        cdef np.ndarray arr = np.empty(self._node_count, dtype=np.float64)
        cdef Py_ssize_t i
        for i in range(self._node_count):
            arr[i] = self._impurity[i]
        return arr

    cpdef np.ndarray predict(self, object X):
        cdef np.ndarray apply = self.apply(X)
        cdef np.ndarray predict = np.take(self.value, apply, axis=0, mode="clip")
        if self._n_labels == 1:
            predict = predict.reshape(X.shape[0])
        return predict

    cpdef np.ndarray apply(self, object X):
        if not isinstance(X, np.ndarray):
            raise ValueError(f"X should be np.ndarray, got {type(X)}")

        cdef TSDatabase ts = ts_database_new(X)
        cdef np.ndarray[np.npy_intp] out = np.zeros((ts.n_samples,), dtype=np.intp)
        cdef Py_ssize_t *out_data = <Py_ssize_t*> out.data
        cdef TSCopy *shapelet
        cdef double threshold
        cdef Py_ssize_t node_index
        cdef Py_ssize_t i
        with nogil:
            for i in range(ts.n_samples):
                node_index = 0
                while self._left[node_index] != -1:
                    threshold = self._thresholds[node_index]
                    shapelet = self._shapelets[node_index]
                    if self.distance_measure.ts_copy_sub_distance(shapelet, &ts, i) <= threshold:
                        node_index = self._left[node_index]
                    else:
                        node_index = self._right[node_index]
                out_data[i] = <Py_ssize_t> node_index
        return out

    cpdef np.ndarray decision_path(self, object X):
        if not isinstance(X, np.ndarray):
            raise ValueError(f"X should be np.ndarray, got {type(X)}")
        cdef TSDatabase ts = ts_database_new(X)
        cdef np.ndarray out = np.zeros((ts.n_samples, self.node_count), order="c", dtype=np.intp)

        cdef Py_ssize_t *out_data = <Py_ssize_t*> out.data
        cdef Py_ssize_t i_stride = <Py_ssize_t> out.strides[0] / <Py_ssize_t> out.itemsize
        cdef Py_ssize_t n_stride = <Py_ssize_t> out.strides[1] / <Py_ssize_t> out.itemsize
        cdef Py_ssize_t node_index
        cdef Py_ssize_t i
        cdef TSCopy *shapelet
        cdef double threshold
        with nogil:
            for i in range(ts.n_samples):
                node_index = 0
                while self._left[node_index] != -1:
                    out_data[i * i_stride + node_index * n_stride] = 1
                    # out_data[i, node_index] = 1
                    threshold = self._thresholds[node_index]
                    shapelet = self._shapelets[node_index]
                    if self.distance_measure.ts_copy_sub_distance(shapelet, &ts, i) <= threshold:
                        node_index = self._left[node_index]
                    else:
                        node_index = self._right[node_index]
        return out

    @property
    def node_count(self):
        return self._node_count

    cdef Py_ssize_t add_leaf_node(
        self,
        Py_ssize_t parent,
        bint is_left,
        Py_ssize_t n_node_samples,
        double n_weighted_node_samples,
    ) nogil:
        cdef Py_ssize_t node_id = self._node_count
        if node_id >= self._capacity:
            if self._increase_capacity() == -1:
                return -1

        self._n_node_samples[node_id] = n_node_samples
        self._n_weighted_node_samples[node_id] = n_weighted_node_samples
        if parent != -1:
            if is_left:
                self._left[parent] = node_id
            else:
                self._right[parent] = node_id
        self._left[node_id] = -1
        self._right[node_id] = -1
        self._impurity[node_id] = -1
        self._shapelets[node_id] = NULL
        self._node_count += 1
        return node_id

    cdef void set_leaf_value(
        self,
        Py_ssize_t node_id,
        Py_ssize_t out_label,
        double out_value,
    ) nogil:
        self._values[out_label + node_id * self._n_labels] = out_value

    cdef Py_ssize_t add_branch_node(
        self,
        Py_ssize_t parent,
        bint is_left,
        Py_ssize_t n_node_samples,
        double n_weighted_node_samples,
        TSCopy *shapelet,
        double threshold,
        double impurity,
    ) nogil:
        cdef Py_ssize_t node_id = self._node_count
        if node_id >= self._capacity:
            if self._increase_capacity() == -1:
                return -1

        self._impurity[node_id] = impurity
        self._n_node_samples[node_id] = n_node_samples
        self._n_weighted_node_samples[node_id] = n_weighted_node_samples
        self._thresholds[node_id] = threshold
        self._shapelets[node_id] = shapelet
        if parent != -1:
            if is_left:
                self._left[parent] = node_id
            else:
                self._right[parent] = node_id

        self._node_count += 1
        return node_id

    cdef Py_ssize_t _increase_capacity(self) nogil except -1:
        cdef Py_ssize_t new_capacity = self._node_count * 2
        cdef Py_ssize_t ret
        ret = safe_realloc(<void**> &self._shapelets, sizeof(TSCopy) * new_capacity)
        if ret == -1:
            return -1

        ret = safe_realloc(<void**> &self._thresholds, sizeof(double) * new_capacity)
        if ret == -1:
            return -1

        ret = safe_realloc(<void**> &self._impurity, sizeof(double) * new_capacity)
        if ret == -1:
            return -1

        ret = safe_realloc(<void**> &self._n_node_samples, sizeof(double) * new_capacity)
        if ret == -1:
            return -1

        ret = safe_realloc(<void**> &self._n_weighted_node_samples, sizeof(Py_ssize_t) * new_capacity)
        if ret == -1:
            return -1

        ret = safe_realloc(<void**> &self._values, sizeof(double) * self._n_labels * new_capacity)
        if ret == -1:
            return -1

        ret = safe_realloc(<void**> &self._left, sizeof(Py_ssize_t) * new_capacity)
        if ret == -1:
            return -1

        ret = safe_realloc(<void**> &self._right, sizeof(Py_ssize_t) * new_capacity)
        if ret == -1:
            return -1

        return 0


cdef inline SplitPoint new_split_point(
    Py_ssize_t split_point,
    double threshold,
    TSView shapelet_info,
) nogil:
    cdef SplitPoint s
    s.split_point = split_point
    s.threshold = threshold
    s.shapelet_info = shapelet_info
    return s


cdef class ShapeletTreeBuilder:
    cdef size_t random_seed

    cdef Py_ssize_t n_shapelets
    cdef Py_ssize_t min_shapelet_size
    cdef Py_ssize_t max_shapelet_size
    cdef Py_ssize_t max_depth
    cdef Py_ssize_t min_sample_split
    cdef Py_ssize_t current_node_id

    # the stride of the label array
    cdef Py_ssize_t label_stride

    # the weight of the j:th sample
    cdef double *sample_weights

    # the dataset of time series
    cdef TSDatabase td

    # the number of samples with non-zero weight
    cdef Py_ssize_t n_samples

    # buffer of samples from 0, ..., n_samples
    cdef Py_ssize_t *samples

    # temporary buffer of samples from 0, ..., n_samples
    cdef Py_ssize_t *samples_buffer

    # the sum of samples_weights
    cdef double n_weighted_samples

    # temporary buffer for distance computations
    cdef double *distance_buffer

    # the distance measure implementation
    cdef DistanceMeasure distance_measure

    # the tree structure representation
    cdef Tree tree

    def __cinit__(
        self,
        Py_ssize_t n_shapelets,
        Py_ssize_t min_shapelet_size,
        Py_ssize_t max_shapelet_size,
        Py_ssize_t max_depth,
        Py_ssize_t min_sample_split,
        object metric,
        dict metric_params,
        np.ndarray X,
        np.ndarray y,
        np.ndarray sample_weights,
        object random_state,
        *args,
        **kwargs
    ):
        self.random_seed = random_state.randint(0, RAND_R_MAX)
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.max_depth = max_depth
        self.min_sample_split = min_sample_split
        self.current_node_id = 0

        self.td = ts_database_new(X)
        self.distance_measure = new_distance_measure(&self.td, metric, metric_params)
        self.label_stride = <Py_ssize_t> y.strides[0] / <Py_ssize_t> y.itemsize

        self.n_samples = X.shape[0]
        self.samples = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * self.n_samples)
        self.samples_buffer = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * self.n_samples)
        self.distance_buffer = <double*> malloc(sizeof(double) * self.n_samples)

        if (
            self.samples == NULL or
            self.distance_buffer == NULL or
            self.samples_buffer == NULL
        ):
            raise MemoryError()

        cdef Py_ssize_t i
        cdef Py_ssize_t j = 0
        for i in range(self.n_samples):
            if sample_weights is None or sample_weights[i] != 0.0:
                self.samples[j] = i
                j += 1

        self.n_samples = j
        self.n_weighted_samples = 0

        if sample_weights is None:
            self.sample_weights = NULL
        else:
            self.sample_weights = <double*> sample_weights.data

    def __dealloc__(self):
        free(self.samples)
        free(self.samples_buffer)
        free(self.distance_buffer)

    @property
    def tree_(self):
        return self.tree

    cpdef Py_ssize_t build_tree(self):
        cdef Py_ssize_t root_node_id
        cdef Py_ssize_t max_depth = 0
        with nogil:
            root_node_id = self._build_tree(0, self.n_samples, 0, -1, False, &max_depth)

        self.tree._max_depth = max_depth
        return root_node_id

    cdef Py_ssize_t new_leaf_node(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t parent,
        bint is_left,
    ) nogil:
        pass

    cdef Py_ssize_t new_branch_node(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        SplitPoint sp,
        TSCopy *shapelet,
        Py_ssize_t parent,
        bint is_left,
    ) nogil:
        cdef Py_ssize_t node_id
        node_id = self.tree.add_branch_node(
            parent,
            is_left,
            end - start,
            self.n_weighted_samples,
            shapelet,
            sp.threshold, -1,
        )
        return node_id


    cdef bint _is_pre_pruned(self, Py_ssize_t start, Py_ssize_t end) nogil:
        """Check if the tree should be pruned based on the samples in the
        region indicated by `start` and `end`

        Implementation detail: For optimization, subclasses *must* set
        the variable `self.n_weighted_samples` to the number of samples
        reaching this node in this method.
       
        Parameters
        ----------
        start : int
            The start position in samples
            
        end : int
            The end position in samples
        """
        pass

    cdef Py_ssize_t _build_tree(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t depth,
        Py_ssize_t parent,
        bint is_left,
        Py_ssize_t *max_depth,
    ) nogil:
        """Recursive function for building the tree

        Each call to this function is allowed to access and transpose
        samples (in `self.samples`) in the region indicated by `start`
        and `end`.

        Parameters
        ----------
        start : int
            The start position in samples
            
        end : int
            The end position in samples
            
        depth : int (out)
            The current recusion depth
            
        parent : int
            The index of the parent node
            
        is_left : bool
            True if current node is left
            
        max_depth : int (out)
            The max depth reached in the tree        
        """
        if depth > max_depth[0]:
            max_depth[0] = depth

        if self._is_pre_pruned(start, end) or depth >= self.max_depth:
            return self.new_leaf_node(start, end, parent, is_left)

        cdef SplitPoint split = self._split(start, end)
        cdef TSCopy *shapelet
        cdef Py_ssize_t current_node_id, left_node_id, right_node_id
        cdef Py_ssize_t err
        if split.split_point > start and end - split.split_point > 0:
            shapelet = <TSCopy*> malloc(sizeof(TSCopy))
            err = self.distance_measure.init_ts_copy(
                shapelet, &split.shapelet_info, &self.td
            )
            if err == -1:
                return -1

            # `shapelet` will be stored and freed by `self.tree`
            current_node_id = self.new_branch_node(
                start, end, split, shapelet, parent, is_left
            )
            left_node_id = self._build_tree(
                start, split.split_point, depth + 1, current_node_id, True, max_depth
            )
            right_node_id = self._build_tree(
                split.split_point, end, depth + 1, current_node_id, False, max_depth
            )

            ts_view_free(&split.shapelet_info)  # RECLAIM THIS MEMORY
            return current_node_id
        else:
            with gil:
                print("warn: split point outside allowed range. This is a bug. Please report me.")
            return self.new_leaf_node(start, end, parent, is_left)

    cdef SplitPoint _split(self, Py_ssize_t start, Py_ssize_t end) nogil:
        """Split the `sample` array indicated by the region specified by
        `start` and `end` by minimizing an impurity measure

        Parameters
        ----------
        start : int
            The start position in samples
            
        end : int
            The end position in samples
            
        Notes
        -----
         - `self._partition_distance_buffer`: implements the decision 
            to determine the split quality minimizing `impurity`
        """
        cdef Py_ssize_t split_point, best_split_point
        cdef double threshold, best_threshold
        cdef double impurity
        cdef double best_impurity
        cdef TSView shapelet
        cdef TSView best_shapelet
        cdef Py_ssize_t i

        ts_view_init(&best_shapelet)
        best_impurity = INFINITY
        best_threshold = NAN
        best_split_point = 0
        split_point = 0

        for i in range(self.n_shapelets):
            self._sample_shapelet(&shapelet, start, end)
            self.distance_measure.ts_view_sub_distances(
                &shapelet,
                &self.td,
                self.samples + start,
                self.distance_buffer + start,
                end - start,
            )
            argsort(self.distance_buffer + start, self.samples + start, end - start)
            self._partition_distance_buffer(
                start, end, &split_point, &threshold, &impurity
            )
            if impurity < best_impurity:
                # store the order of samples in `sample_buffer`
                memcpy(
                    self.samples_buffer,
                    self.samples + start,
                    sizeof(Py_ssize_t) * (end - start),
                )
                best_impurity = impurity
                best_split_point = split_point
                best_threshold = threshold
                best_shapelet = shapelet
            else:
                ts_view_free(&shapelet)

        # restore the best order to `samples`
        memcpy(
            self.samples + start,
            self.samples_buffer,
            sizeof(Py_ssize_t) * (end - start),
        )
        return new_split_point(best_split_point, best_threshold, best_shapelet)

    cdef Py_ssize_t _sample_shapelet(
        self,
        TSView *shapelet_info,
        Py_ssize_t start,
        Py_ssize_t end,
    ) nogil:
        cdef Py_ssize_t shapelet_length
        cdef Py_ssize_t shapelet_start
        cdef Py_ssize_t shapelet_index
        cdef Py_ssize_t shapelet_dim

        shapelet_length = rand_int(self.min_shapelet_size,
                                   self.max_shapelet_size,
                                   &self.random_seed)
        shapelet_start = rand_int(0, self.td.n_timestep - shapelet_length,
                                  &self.random_seed)
        shapelet_index = self.samples[rand_int(start, end, &self.random_seed)]
        if self.td.n_dims > 1:
            shapelet_dim = rand_int(0, self.td.n_dims, &self.random_seed)
        else:
            shapelet_dim = 1
        return self.distance_measure.init_ts_view(&self.td,
                                                  shapelet_info,
                                                  shapelet_index,
                                                  shapelet_start,
                                                  shapelet_length,
                                                  shapelet_dim)

    cdef void _partition_distance_buffer(self,
                                         Py_ssize_t start,
                                         Py_ssize_t end,
                                         Py_ssize_t *split_point,
                                         double *threshold,
                                         double *impurity) nogil:
        """ Partition the distance buffer such that an impurity measure is
        minimized
        
        Parameters
        ----------
        start : int
            The start position in samples
            
        end : int
            The end position in samples
            
        split_point : int (out) 
            The split point: an index in `range(start, end)` minimizing impurity

        threshold : float (out) 
            The threshold value

        impurity : float (in, out) 
            - The initial impurity (in) 
            - The optimal impurity (out)
            
        Warnings
        --------
        - Must be implemented by subclasses
        """
        with gil:
            raise NotImplementedError()


cdef class ClassificationShapeletTreeBuilder(ShapeletTreeBuilder):
    # the number of labels
    cdef Py_ssize_t n_labels

    # temporary buffer with sample distributions per label
    cdef double *label_buffer

    # temporary buffer with the left split sample distributions
    cdef double *left_label_buffer

    # temporary buffer with the right split sample distribution
    cdef double *right_label_buffer

    # the (strided) array of labels
    cdef Py_ssize_t *labels

    def __cinit__(
        self,
        Py_ssize_t n_shapelets,
        Py_ssize_t min_shapelet_size,
        Py_ssize_t max_shapelet_size,
        Py_ssize_t max_depth,
        Py_ssize_t min_sample_split,
        object metric,
        object metric_params,
        np.ndarray X,
        np.ndarray y,
        np.ndarray sample_weights,
        object random_state,
        Py_ssize_t n_labels
    ):
        self.labels = <Py_ssize_t*> y.data
        self.n_labels = n_labels
        self.tree = Tree(self.distance_measure, n_labels)
        self.label_buffer = <double*> malloc(sizeof(double) * n_labels)
        self.left_label_buffer = <double*> malloc(sizeof(double) * n_labels)
        self.right_label_buffer = <double*> malloc(sizeof(double) * n_labels)
        if (self.left_label_buffer == NULL or
                self.right_label_buffer == NULL or
                self.label_buffer == NULL):
            raise MemoryError()

    def __dealloc__(self):
        free(self.label_buffer)
        free(self.left_label_buffer)
        free(self.right_label_buffer)

    cdef Py_ssize_t new_leaf_node(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t parent,
        bint is_left,
    ) nogil:
        cdef Py_ssize_t node_id
        cdef Py_ssize_t i
        cdef double prob

        node_id = self.tree.add_leaf_node(
            parent, is_left, end - start, self.n_weighted_samples
        )
        for i in range(self.n_labels):
            prob = self.label_buffer[i] / self.n_weighted_samples
            self.tree.set_leaf_value(node_id, i, prob)
        return node_id

    cdef bint _is_pre_pruned(self, Py_ssize_t start, Py_ssize_t end) nogil:
        # reinitialize the `label_buffer` to the sample distribution
        # in the current sample region.
        memset(self.label_buffer, 0, sizeof(double) * self.n_labels)
        cdef Py_ssize_t n_positive = label_distribution(
            self.samples,
            self.sample_weights,
            start,
            end,
            self.labels,
            self.label_stride,
            self.n_labels,
            &self.n_weighted_samples,  # out param
            self.label_buffer,  # out param
        )

        if end - start <= self.min_sample_split or n_positive < 2:
            return True
        return False

    cdef void _partition_distance_buffer(
            self,
            Py_ssize_t start,
            Py_ssize_t end,
            Py_ssize_t *split_point,
            double *threshold,
            double *impurity,
    ) nogil:
        memset(self.left_label_buffer, 0, sizeof(double) * self.n_labels)

        # store the label buffer temporarily in `right_label_buffer`,
        # since all samples fall on the right hand side of the threshold
        memcpy(self.right_label_buffer, self.label_buffer,
               sizeof(double) * self.n_labels)

        cdef Py_ssize_t i  # real index of samples (in `range(start, end)`)
        cdef Py_ssize_t j  # sample index (in `samples`)
        cdef Py_ssize_t p  # label index (in `label_buffer`)

        cdef double right_sum
        cdef double left_sum

        cdef double prev_distance
        cdef Py_ssize_t prev_label

        cdef double current_sample_weight
        cdef double current_distance
        cdef double current_impurity
        cdef Py_ssize_t current_label

        j = self.samples[start]
        p = j * self.label_stride

        prev_distance = self.distance_buffer[start]
        prev_label = self.labels[p]

        if self.sample_weights != NULL:
            current_sample_weight = self.sample_weights[j]
        else:
            current_sample_weight = 1.0

        left_sum = current_sample_weight
        right_sum = self.n_weighted_samples - current_sample_weight

        self.left_label_buffer[prev_label] += current_sample_weight
        self.right_label_buffer[prev_label] -= current_sample_weight

        impurity[0] = entropy(left_sum,
                              self.left_label_buffer,
                              right_sum,
                              self.right_label_buffer,
                              self.n_labels)

        threshold[0] = prev_distance
        # The split point indicates a <=-relation
        split_point[0] = start + 1

        for i in range(start + 1, end - 1):
            j = self.samples[i]
            current_distance = self.distance_buffer[i]

            p = j * self.label_stride
            current_label = self.labels[p]

            if not current_label == prev_label:
                current_impurity = entropy(
                    left_sum,
                    self.left_label_buffer,
                    right_sum,
                    self.right_label_buffer,
                    self.n_labels,
                )

                if current_impurity <= impurity[0]:
                    impurity[0] = current_impurity
                    threshold[0] = (current_distance + prev_distance) / 2
                    split_point[0] = i

            if self.sample_weights != NULL:
                current_sample_weight = self.sample_weights[j]
            else:
                current_sample_weight = 1.0

            left_sum += current_sample_weight
            right_sum -= current_sample_weight
            self.left_label_buffer[current_label] += current_sample_weight
            self.right_label_buffer[current_label] -= current_sample_weight

            prev_label = current_label
            prev_distance = current_distance


cdef class RegressionShapeletTreeBuilder(ShapeletTreeBuilder):
    # the (strided) array of labels
    cdef double *labels
    cdef RollingVariance right
    cdef RollingVariance left

    def __cinit__(
        self,
        Py_ssize_t n_shapelets,
        Py_ssize_t min_shapelet_size,
        Py_ssize_t max_shapelet_size,
        Py_ssize_t max_depth,
        Py_ssize_t min_sample_split,
        object metric,
        object metric_params,
        np.ndarray X,
        np.ndarray y,
        np.ndarray sample_weights,
        object random_state,
    ):
        self.labels = <double*> y.data
        self.tree = Tree(self.distance_measure, 1)
        self.left = RollingVariance()
        self.right = RollingVariance()

    cdef bint _is_pre_pruned(self, Py_ssize_t start, Py_ssize_t end) nogil:
        cdef Py_ssize_t j
        cdef double sample_weight
        self.n_weighted_samples = 0
        for i in range(start, end):
            j = self.samples[i]
            if self.sample_weights != NULL:
                sample_weight = self.sample_weights[j]
            else:
                sample_weight = 1.0

            self.n_weighted_samples += sample_weight

        if end - start <= self.min_sample_split:
            return True
        return False

    cdef Py_ssize_t new_leaf_node(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t parent,
        bint is_left,
    ) nogil:
        cdef double leaf_sum = 0
        cdef double current_sample_weight = 0
        cdef Py_ssize_t j
        cdef Py_ssize_t node_id
        for i in range(start, end):
            j = self.samples[i]
            p = j * self.label_stride
            if self.sample_weights != NULL:
                current_sample_weight = self.sample_weights[j]
            else:
                current_sample_weight = 1.0
            leaf_sum += self.labels[p] * current_sample_weight

        node_id = self.tree.add_leaf_node(parent, is_left, end - start, self.n_weighted_samples)
        self.tree.set_leaf_value(node_id, 0, leaf_sum / self.n_weighted_samples)
        return node_id

    cdef void _partition_distance_buffer(self,
                                         Py_ssize_t start,
                                         Py_ssize_t end,
                                         Py_ssize_t *split_point,
                                         double *threshold,
                                         double *impurity) nogil:
        """Partitions the distance buffer into two binary partitions
        such that the sum of label variance in the two partitions is
        minimized.
       
        The implementation uses an efficient one-pass algorithm [1]
        for computing the variance of the two partitions and finding
        the optimal split point
        
        References
        ----------
        West, D. H. D. (1979). 
            Updating Mean and Variance Estimates: An Improved Method
        """
        cdef Py_ssize_t i  # real index of samples (in `range(start, end)`)
        cdef Py_ssize_t j  # sample index (in `samples`)
        cdef Py_ssize_t p  # label index (in `labels`)

        cdef double prev_distance

        cdef double current_sample_weight
        cdef double current_distance
        cdef double current_impurity
        cdef double current_val

        j = self.samples[start]
        p = j * self.label_stride

        prev_distance = self.distance_buffer[start]
        current_val = self.labels[p]

        if self.sample_weights != NULL:
            current_sample_weight = self.sample_weights[j]
        else:
            current_sample_weight = 1.0

        self.left._reset()
        self.right._reset()
        self.left._add(current_sample_weight, current_val)

        for i in range(start + 1, end):
            j = self.samples[i]
            p = j * self.label_stride
            if self.sample_weights != NULL:
                current_sample_weight = self.sample_weights[j]
            else:
                current_sample_weight = 1.0

            current_val = self.labels[p]
            self.right._add(current_sample_weight, current_val)

        impurity[0] = self.left._variance() + self.right._variance()
        threshold[0] = prev_distance
        split_point[0] = start + 1  # The split point indicates a <=-relation

        for i in range(start + 1, end - 1):
            j = self.samples[i]
            p = j * self.label_stride

            current_distance = self.distance_buffer[i]

            if self.sample_weights != NULL:
                current_sample_weight = self.sample_weights[j]
            else:
                current_sample_weight = 1.0

            # Move variance from the right-hand side to the left-hand
            # and reevaluate the impurity
            current_val = self.labels[p]
            self.right._remove(current_sample_weight, current_val)
            self.left._add(current_sample_weight, current_val)

            current_impurity = self.left._variance() + self.right._variance()
            if current_impurity <= impurity[0]:
                impurity[0] = current_impurity
                threshold[0] = (current_distance + prev_distance) / 2
                split_point[0] = i

            prev_distance = current_distance


cdef class ExtraRegressionShapeletTreeBuilder(RegressionShapeletTreeBuilder):

    cdef void _partition_distance_buffer(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t *split_point,
        double *threshold,
        double *impurity,
    ) nogil:
        # The smallest distance is always 0
        cdef double min_dist = self.distance_buffer[start + 1]
        cdef double max_dist = self.distance_buffer[end - 1]
        cdef double rand_threshold = rand_uniform(
            min_dist, max_dist, &self.random_seed
        )
        cdef Py_ssize_t i
        split_point[0] = start + 1
        for i in range(start + 1, end - 1):
            if self.distance_buffer[i] <= rand_threshold:
                split_point[0] = i
            else:
                break
        threshold[0] = rand_threshold
        # TODO: compute impurity scoring
        impurity[0] = 0


cdef class ExtraClassificationShapeletTreeBuilder(ClassificationShapeletTreeBuilder):

    cdef void _partition_distance_buffer(
        self,
        Py_ssize_t start,
        Py_ssize_t end,
        Py_ssize_t *split_point,
        double *threshold,
        double *impurity,
    ) nogil:
        # The smallest distance is always 0
        cdef double min_dist = self.distance_buffer[start + 1]
        cdef double max_dist = self.distance_buffer[end - 1]
        cdef double rand_threshold = rand_uniform(
            min_dist, max_dist, &self.random_seed
        )
        cdef Py_ssize_t i

        split_point[0] = start + 1
        for i in range(start + 1, end - 1):
            if self.distance_buffer[i] <= rand_threshold:
                split_point[0] = i
            else:
                break
        threshold[0] = rand_threshold
        # TODO: compute impurity scoring
        impurity[0] = -INFINITY



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

import numpy as np
cimport numpy as np

from libc.stdlib cimport malloc
from libc.stdlib cimport free

from libc.math cimport sqrt
from libc.math cimport INFINITY

from ._distance cimport ScaledDistanceMeasure
from ._distance cimport DistanceMeasure

from ._distance cimport TSDatabase

from ._distance cimport TSCopy
from ._distance cimport TSView

from .._utils cimport realloc_array


cdef class ScaledEuclideanDistance(ScaledDistanceMeasure):
    cdef double *X_buffer


    def __cinit__(self, Py_ssize_t n_timestep, *args, **kwargs):
        super().__init__(n_timestep)
        self.X_buffer = <double*> malloc(sizeof(double) * n_timestep * 2)


    cdef double ts_copy_sub_distance(
        self,
        TSCopy *ts_copy,
        TSDatabase *td,
        Py_ssize_t t_index,
        Py_ssize_t *return_index = NULL,
    ) nogil:
        cdef Py_ssize_t sample_offset = (
            t_index * td.sample_stride + ts_copy.dim * td.dim_stride
        )

        return scaled_euclidean_distance(
            0,
            1,
            ts_copy.length,
            ts_copy.mean,
            ts_copy.std,
            ts_copy.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            self.X_buffer,
            return_index,
        )


    def __dealloc__(self):
        free(self.X_buffer)


    cdef double ts_view_sub_distance(
        self,
        TSView *ts_view,
        TSDatabase *td,
        Py_ssize_t t_index,
    ) nogil:
        cdef Py_ssize_t sample_offset = (
            t_index * td.sample_stride + ts_view.dim * td.dim_stride
        )
        cdef Py_ssize_t shapelet_offset = (
            ts_view.index * td.sample_stride +
            ts_view.dim * td.dim_stride +
            ts_view.start * td.timestep_stride
        )
        return scaled_euclidean_distance(
            shapelet_offset,
            td.timestep_stride,
            ts_view.length,
            ts_view.mean,
            ts_view.std,
            td.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            self.X_buffer,
            NULL,
        )


    cdef int ts_copy_matches(
        self,
        TSCopy *ts_copy,
        TSDatabase *td,
        Py_ssize_t t_index,
        double threshold,
        Py_ssize_t** matches,
        double** distances,
        Py_ssize_t *n_matches,
    ) nogil except -1:
        cdef Py_ssize_t sample_offset = (
            t_index * td.sample_stride + ts_copy.dim * td.dim_stride
        )

        return scaled_euclidean_distance_matches(
            0,
            1,
            ts_copy.length,
            ts_copy.mean,
            ts_copy.std,
            ts_copy.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            self.X_buffer,
            threshold,
            distances,
            matches,
            n_matches,
        )


cdef class EuclideanDistance(DistanceMeasure):

    cdef double ts_copy_sub_distance(
        self,
        TSCopy *ts_copy,
        TSDatabase *td,
        Py_ssize_t t_index,
        Py_ssize_t *return_index = NULL,
    ) nogil:
        cdef Py_ssize_t sample_offset = (
            t_index * td.sample_stride + ts_copy.dim * td.dim_stride
        )

        return euclidean_distance(
            0,
            1,
            ts_copy.length,
            ts_copy.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            return_index,
        )


    cdef double ts_view_sub_distance(
        self,
        TSView *ts_view,
        TSDatabase *td,
        Py_ssize_t t_index,
    ) nogil:
        cdef Py_ssize_t sample_offset = (
            t_index * td.sample_stride + ts_view.dim * td.dim_stride
        )
        cdef Py_ssize_t shapelet_offset = (
            ts_view.index * td.sample_stride +
            ts_view.dim * td.dim_stride +
            ts_view.start * td.timestep_stride
        )
        return euclidean_distance(
            shapelet_offset,
            td.timestep_stride,
            ts_view.length,
            td.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            NULL,
        )


    cdef int ts_copy_matches(
        self,
        TSCopy *ts_copy,
        TSDatabase *td,
        Py_ssize_t t_index,
        double threshold,
        Py_ssize_t** matches,
        double** distances,
        Py_ssize_t *n_matches,
    ) nogil except -1:
        cdef Py_ssize_t sample_offset = (
            t_index * td.sample_stride + ts_copy.dim * td.dim_stride
        )
        return euclidean_distance_matches(
            0,
            1,
            ts_copy.length,
            ts_copy.data,
            sample_offset,
            td.timestep_stride,
            td.n_timestep,
            td.data,
            threshold,
            distances,
            matches,
            n_matches,
        )


cdef double scaled_euclidean_distance(
    Py_ssize_t s_offset,
    Py_ssize_t s_stride,
    Py_ssize_t s_length,
    double s_mean,
    double s_std,
    double *S,
    Py_ssize_t t_offset,
    Py_ssize_t t_stride,
    Py_ssize_t t_length,
    double *T,
    double *X_buffer,
    Py_ssize_t *index,
) nogil:
    cdef double current_value = 0
    cdef double mean = 0
    cdef double std = 0
    cdef double dist = 0
    cdef double min_dist = INFINITY

    cdef double ex = 0
    cdef double ex2 = 0

    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef Py_ssize_t buffer_pos

    for i in range(t_length):
        current_value = T[t_offset + t_stride * i]
        ex += current_value
        ex2 += current_value * current_value

        buffer_pos = i % s_length
        X_buffer[buffer_pos] = current_value
        X_buffer[buffer_pos + s_length] = current_value
        if i >= s_length - 1:
            j = (i + 1) % s_length
            mean = ex / s_length
            std = sqrt(ex2 / s_length - mean * mean)
            dist = inner_scaled_euclidean_distance(s_offset, s_length, s_mean, s_std,
                                                   j, mean, std, S, s_stride,
                                                   X_buffer, min_dist)

            if dist < min_dist:
                min_dist = dist
                if index != NULL:
                    index[0] = (i + 1) - s_length

            current_value = X_buffer[j]
            ex -= current_value
            ex2 -= current_value * current_value

    return sqrt(min_dist)


cdef double inner_scaled_euclidean_distance(
    Py_ssize_t offset,
    Py_ssize_t length,
    double s_mean,
    double s_std,
    Py_ssize_t j,
    double mean,
    double std,
    double *X,
    Py_ssize_t timestep_stride,
    double *X_buffer,
    double min_dist,
) nogil:
    # Compute the distance between the shapelet (starting at `offset`
    # and ending at `offset + length` normalized with `s_mean` and
    # `s_std` with the shapelet in `X_buffer` starting at `0` and
    # ending at `length` normalized with `mean` and `std`
    cdef double dist = 0
    cdef double x
    cdef Py_ssize_t i
    cdef bint std_zero = std == 0
    cdef bint s_std_zero = s_std == 0

    # distance is zero
    if s_std_zero and std_zero:
        return 0

    for i in range(length):
        if dist >= min_dist:
            break

        x = (X[offset + timestep_stride * i] - s_mean) / s_std
        if not std_zero:
            x -= (X_buffer[i + j] - mean) / std
        dist += x * x

    return dist


cdef double euclidean_distance(
    Py_ssize_t s_offset,
    Py_ssize_t s_stride,
    Py_ssize_t s_length,
    double *S,
    Py_ssize_t t_offset,
    Py_ssize_t t_stride,
    Py_ssize_t t_length,
    double *T,
    Py_ssize_t *index,
) nogil:
    cdef double dist = 0
    cdef double min_dist = INFINITY

    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef double x
    for i in range(t_length - s_length + 1):
        dist = 0
        for j in range(s_length):
            if dist >= min_dist:
                break

            x = T[t_offset + t_stride * i + j]
            x -= S[s_offset + s_stride * j]
            dist += x * x

        if dist < min_dist:
            min_dist = dist
            if index != NULL:
                index[0] = i

    return sqrt(min_dist)


cdef int euclidean_distance_matches(
    Py_ssize_t s_offset,
    Py_ssize_t s_stride,
    Py_ssize_t s_length,
    double *S,
    Py_ssize_t t_offset,
    Py_ssize_t t_stride,
    Py_ssize_t t_length,
    double *T,
    double threshold,
    double **distances,
    Py_ssize_t **matches,
    Py_ssize_t *n_matches,
) nogil except -1:
    cdef double dist = 0
    cdef Py_ssize_t capacity = 1
    cdef Py_ssize_t tmp_capacity
    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef double x

    matches[0] = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * capacity)
    distances[0] = <double*> malloc(sizeof(double) * capacity)
    n_matches[0] = 0

    threshold = threshold * threshold
    for i in range(t_length - s_length + 1):
        dist = 0
        for j in range(s_length):
            if dist > threshold:
                break

            x = T[t_offset + t_stride * i + j]
            x -= S[s_offset + s_stride * j]
            dist += x * x
        if dist <= threshold:
            tmp_capacity = capacity
            realloc_array(<void**> matches, n_matches[0], sizeof(Py_ssize_t), &tmp_capacity)
            realloc_array(<void**> distances, n_matches[0], sizeof(double), &capacity)
            matches[0][n_matches[0]] = i
            distances[0][n_matches[0]] = sqrt(dist)
            n_matches[0] += 1

    return 0


cdef int scaled_euclidean_distance_matches(
   Py_ssize_t s_offset,
   Py_ssize_t s_stride,
   Py_ssize_t s_length,
   double s_mean,
   double s_std,
   double *S,
   Py_ssize_t t_offset,
   Py_ssize_t t_stride,
   Py_ssize_t t_length,
   double *T,
   double *X_buffer,
   double threshold,
   double** distances,
   Py_ssize_t** matches,
   Py_ssize_t *n_matches,
) nogil except -1:
    cdef double current_value = 0
    cdef double mean = 0
    cdef double std = 0
    cdef double dist = 0

    cdef double ex = 0
    cdef double ex2 = 0

    cdef Py_ssize_t i
    cdef Py_ssize_t j
    cdef Py_ssize_t buffer_pos
    cdef Py_ssize_t capacity = 1
    cdef Py_ssize_t tmp_capacity

    matches[0] = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * capacity)
    distances[0] = <double*> malloc(sizeof(double) * capacity)
    n_matches[0] = 0

    threshold = threshold * threshold

    for i in range(t_length):
        current_value = T[t_offset + t_stride * i]
        ex += current_value
        ex2 += current_value * current_value

        buffer_pos = i % s_length
        X_buffer[buffer_pos] = current_value
        X_buffer[buffer_pos + s_length] = current_value
        if i >= s_length - 1:
            j = (i + 1) % s_length
            mean = ex / s_length
            std = sqrt(ex2 / s_length - mean * mean)
            dist = inner_scaled_euclidean_distance(
                s_offset, s_length, s_mean, s_std, j, mean, std, S, s_stride,
                X_buffer, threshold)

            if dist <= threshold:
                tmp_capacity = capacity
                realloc_array(
                    <void**> matches, n_matches[0], sizeof(Py_ssize_t), &tmp_capacity)
                realloc_array(
                    <void**> distances, n_matches[0], sizeof(double), &capacity)

                matches[0][n_matches[0]] = (i + 1) - s_length
                distances[0][n_matches[0]] = sqrt(dist)

                n_matches[0] += 1

            current_value = X_buffer[j]
            ex -= current_value
            ex2 -= current_value * current_value

    return 0



import math

import numpy as np

from wildboar._utils import check_array_fast
from wildboar.distance import distance
from wildboar.distance.dtw import (
    dtw_pairwise_distance,
    dtw_alignment,
    dtw_distance,
    dtw_envelop,
    dtw_lb_keogh,
)

from wildboar.datasets import load_two_lead_ecg, load_dataset

import matplotlib.pylab as plt

# x, y = load_two_lead_ecg()
# d, i = distance(x[0, 10:20], x, sample=[0, 1, 2, 3, 5, 10], metric='scaled_dtw', metric_params={'r': 0}, return_index=True)
# print(d)
# print(i)
#
# d, i = matches(x[0, 10:20], x, 1.7, sample=[0, 1, 2], metric='scaled_euclidean', metric_params={'r': 0}, return_distance=True)
# print(d)
# print(i)

# print("DIST 1")
# print(distance([1, 2, 3], x, dim=0))
check_array_fast(np.array([1, 2, 3, 2]).reshape(2, 2), ensure_2d=False, allow_nd=True)
x, y = load_dataset("GunPoint")
# rnd = np.random.RandomState(123)
# x = rnd.randn(100)
# y = rnd.randn(100)

l, u = dtw_envelop(x[-1], r=10)
lu = dtw_lb_keogh(x[-1], x[0], r=10)
plt.plot(x[-1])
plt.plot(x[0])
plt.plot(l, "b--")
plt.plot(u, "r--")
plt.plot(lu, "g--")
plt.show()

x = np.array([[1, 10, 2, 3], [1, 2, 3, 4], [2, 1, 2, 2]])
#
print(
    distance(
        [1, 2, 3], x, metric="scaled_dtw", metric_params={"r": 0}, return_index=True
    )
)
# print(np.linalg.norm(s - x))
#
# x = np.random.randn(100, 100)
print(dtw_pairwise_distance(x, r=1))
# t = np.array([-1, -2, -3, -4, 10, 1, 1, 2], dtype=float)
# s, t = t, s

# s = (s - np.mean(s)) / np.std(s)
# t = (t - np.mean(t)) / np.std(t)
# print(s)
# print(t)

# alignment = dtw_alignment(s, t, r=1, out=np.zeros((s.shape[0], t.shape[0])))
# print(np.sqrt(alignment[-1, -1]))
#
# s_scaled = (s - np.mean(s)) / np.std(s)
# t_scaled = (t - np.mean(t)) / np.std(t)

# print(distance(s, t, metric="dtw", metric_params={"r": 1}, subsequence_distance=False))
# print(
#     distance(
#         s, t, metric="scaled_dtw", metric_params={"r": 2}, subsequence_distance=False
#     )
# )
# print(
#     distance(
#         s_scaled,
#         t_scaled,
#         metric="dtw",
#         metric_params={"r": 2},
#         subsequence_distance=False,
#     )
# )
# print(dtw_distance(s, t, r=1))
# print(dtw_distance(s, t, r=2, scale=True))
# print(dtw_distance(s_scaled, t_scaled, r=2, scale=False))
# print(np.sqrt(dtw_alignment(s, t, r=2)[-1, -1]))


# x = np.random.randn(1000000)
# y = np.random.randn(1000000)
# print(distance(x, y, metric="dtw", metric_params={"r": 2}, subsequence_distance=False))
# print(dtw_distance(x, y, r=1))
# import timeit
#
# print(timeit.timeit(lambda: dtw_distance(x, y, r=1), number=10))
# print(
#     timeit.timeit(
#         lambda: distance(
#             x, y, metric="dtw", metric_params={"r": 1}, subsequence_distance=False
#         ),
#         number=10,
#     )
# )
# print(timeit.timeit(lambda: np.linalg.norm(x - y), number=10))

# from wildboar.distance import matches

# from scipy.stats import norm

# n_samples = 1000
# n_features = 10000
# n_classes = 2
#
# rng = np.random.RandomState(41)
# np.zeros([10, 10])
#
# delta = 0.5
# dt = 1
#
# X = (norm.rvs(
#     scale=delta**2 * dt,
#     size=n_samples * n_features,
#     random_state=rng,
# ).reshape((n_samples, n_features)))
#
# x = X[0, :]
# data = X[1:, :]
#
# d, i = distance(
#     x[0:10],
#     data,
#     dim=0,
#     metric="scaled_dtw",
#     metric_params={"r": 3},
#     sample=None,
#     return_index=True,
# )
#
# d, i = matches(
#     x[0:10],
#     data,
#     0.37,
#     dim=0,
#     metric="euclidean",
#     sample=10,
#     return_distance=True,
# )
#
# print(d)
# print(i)
#



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

from logging import log
import os

SEM_VER_REGEX = "^v(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"

from pkg_resources import get_distribution, parse_version

release = get_distribution("wildboar").version
version = ".".join(release.split(".")[:3])
# -- Project information -----------------------------------------------------

project = "Wildboar"
description = "Time series learning with Python"
copyright = "2020, Isak Samsten"
author = "Isak Samsten"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "autoapi.extension",
    "sphinx_multiversion",
    "sphinx_panels",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

autoapi_dirs = "autoapi"
autoapi_root = "."
autoapi_template_dir = "_templates/autoapi/"
autoapi_ignore = ["*tests*", "_*.py"]
autoapi_keep_files = True
autoapi_add_toctree_entry = False


autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
    "imported-members",
]
# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
html_theme_options = {}
html_sidebars = {
    "**": [
        "sidebar/scroll-start.html",
        "brand.html",
        "sidebar/search.html",
        "sidebar/navigation.html",
        "versions.html",
        "sidebar/scroll-end.html",
    ],
}
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
smv_branch_whitelist = r"^master|\d+\.\d+\.X$"
smv_remote_whitelist = r"^origin$"
smv_released_pattern = r"^refs/(heads|remotes)/\d+\.\d+\.X$"
smv_tag_whitelist = None


def setup(app):
    import re
    import pathlib
    from sphinx.util.logging import getLogger

    logger = getLogger(__name__)

    def is_tag_for(gitref, major, minor):
        match = re.match(SEM_VER_REGEX, gitref.name)
        return (
            gitref.source == "tags"
            and match
            and match.group(1) == major
            and match.group(2) == minor
        )

    def read_latest_version(app, config):
        if not hasattr(config, "smv_metadata") or len(config.smv_metadata) == 0:
            return

        logger.info("[CONF] installed version %s", config.release)

        from sphinx_multiversion import git

        gitroot = pathlib.Path(
            git.get_toplevel_path(cwd=os.path.abspath(app.confdir))
        ).resolve()
        gitrefs = list(git.get_all_refs(gitroot))
        latest_version_tags = {}
        for ver, metadata in config.smv_metadata.items():
            current_version = re.match("(\d)\.(\d)(?:.X)?", ver)
            if current_version:
                major = current_version.group(1)
                minor = current_version.group(2)
                matching_tags = [
                    re.sub("^v", "", gitref.name)
                    for gitref in gitrefs
                    if is_tag_for(gitref, major, minor)
                ]
                matching_tags.sort(key=parse_version)
                latest_tag = matching_tags[-1] if matching_tags else ver
                latest_version_tags[ver] = latest_tag
            elif ver == "master":
                latest_version_tags[ver] = config.version
            else:
                logger.warning("[CONF] using branch version for tag (%s)", ver)
                latest_version_tags[ver] = ver

            logger.info(
                "[CONF] latest version tag for '%s' is '%s'",
                ver,
                latest_version_tags[ver],
            )
            metadata["version"] = latest_version_tags[ver]

        if config.smv_current_version != "master":
            config.version = config.smv_metadata[config.smv_current_version]["version"]
            config.release = config.version

        config.smv_latest_version_tags = latest_version_tags
        config.smv_current_version_tag = latest_version_tags[config.smv_current_version]

        version_sorted = sorted(
            config.smv_metadata.keys(),
            key=lambda x: parse_version(re.sub(".X$", "", x)),
        )
        config.smv_latest_stable = version_sorted[-1] if version_sorted else "master"

        logger.info("[DOCS] latest stable version is %s", config.smv_latest_stable)

    def set_latest_version(app, pagename, templatename, context, doctree):
        if not hasattr(app.config, "smv_metadata") or len(app.config.smv_metadata) == 0:
            return

        from sphinx_multiversion.sphinx import VersionInfo

        versioninfo = VersionInfo(
            app, context, app.config.smv_metadata, app.config.smv_current_version
        )
        context["current_version_tag"] = app.config.smv_current_version_tag
        context["latest_version_tags"] = app.config.smv_latest_version_tags
        context["latest_stable_version"] = versioninfo[app.config.smv_latest_stable]

    app.connect("config-inited", read_latest_version)
    app.connect("html-page-context", set_latest_version)



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

import numpy as np
from wildboar.datasets._filter import make_str_filter


class TestStrFilter:
    def test_n_samples(self):
        f = make_str_filter("n_samples<100")
        assert f("hello", np.zeros((101, 10)), None) == False

    def test_dataset(self):
        f = make_str_filter("dataset=~TwoLead")
        assert f("TwoLeadECG", None, None) == True



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3

# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Authors: Isak Samsten

from libc.math cimport fabs
from libc.math cimport log2

cdef inline double entropy(double  left_sum,
                           double *left_count,
                           double  right_sum,
                           double *right_count,
                           Py_ssize_t  n_labels) nogil:
    """Compute the entropy impurity

    :param left_sum: number of samples in the left partition
    :param left_count: number of samples per label in the left partition
    :param right_sum: number of samples in the right partiton
    :param right_count: number of samples per label in the right partition
    :param n_labels: the number of labels
    """
    cdef double n_samples = left_sum + right_sum
    cdef double x_sum = 0
    cdef double y_sum = 0
    cdef double xv, yv
    cdef Py_ssize_t i
    for i in range(n_labels):
        xv = left_count[i] / n_samples
        yv = right_count[i] / n_samples
        if xv > 0:
            x_sum += xv * log2(xv)
        if yv > 0:
            y_sum += yv * log2(yv)

    return fabs((left_sum / n_samples) * -x_sum +
                (right_sum / n_samples) * -y_sum)



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

from . import _distance
from . import dtw

__all__ = ["distance", "matches", "dtw"]


def paired_distance(
    x,
    y,
    *,
    metric="euclidean",
    metric_params=None,
    subsequence_distance=True,
):
    raise NotImplementedError()


def distance(
    x,
    y,
    *,
    dim=0,
    sample=None,
    metric="euclidean",
    metric_params=None,
    subsequence_distance=True,
    return_index=False,
):
    """Computes the distance between x and the samples of y

    Parameters
    ----------
    x : array-like of shape (x_timestep, )
        A 1-dimensional float array

    y : array-like of shape (n_samples, n_timesteps) or (n_samples, n_dims, n_timesteps)

    dim : int, optional
        The time series dimension to search

    sample : int or array-like, optional
        The samples to compare to

        - if ``sample=None`` the distances to all samples in data is returned
        - if sample is an int the distance to that sample is returned
        - if sample is an array-like the distance to all samples in sample are returned
        - if ``n_samples=1``, ``samples`` is an int or ``len(samples)==1`` a scalar is returned
        - otherwise an array is returned

    metric : {'euclidean', 'scaled_euclidean', 'dtw', 'scaled_dtw'} or callable, optional
        The distance metric

        - if str use optimized implementations of the named distance measure
        - if callable a function taking two arrays as input

    metric_params: dict, optional
        Parameters to the metric

        - 'euclidean' and 'scaled_euclidean' take no parameters
        - 'dtw' and 'scaled_dtw' take a single paramater 'r'. If 'r' <= 1 it
          is interpreted as a fraction of the time series length. If > 1 it
          is interpreted as an exact time warping window. Use 'r' == 0 for
          a widow size of exactly 1.

    subsequence_distance: bool, optional
        - if True, compute the minimum subsequence distance
        - if False, compute the distance between two arrays of the same length
          unless the specified metric support unaligned arrays

    return_index : bool, optional
        - if True return the index of the best match. If there are many equally good
          matches, the first match is returned.

    Returns
    -------
    dist : float, ndarray
        The smallest distance to each time series

    indices : int, ndarray
        The start position of the best match in each time series

    See Also
    --------
    matches : find shapelets within a threshold

    Examples
    --------

    >>> from wildboar.datasets import load_two_lead_ecg
    >>> x, y = load_two_lead_ecg()
    >>> _, i = distance(x[0, 10:20], x, sample=[0, 1, 2, 3, 5, 10],
    ...                 metric="scaled_euclidean", return_index=True)
    >>> i
    [10 29  9 72 20 30]
    """
    return _distance.distance(
        x,
        y,
        dim=dim,
        sample=sample,
        metric=metric,
        metric_params=metric_params,
        subsequence_distance=subsequence_distance,
        return_index=return_index,
    )


def matches(
    x,
    y,
    threshold,
    *,
    dim=0,
    sample=None,
    metric="euclidean",
    metric_params=None,
    return_distance=False,
):
    """Return the positions in `x` (one list per `sample`) where `x` is closer than `threshold`.

    Parameters
    ----------
    x : array-like of shape (x_timestep, )
        A 1-dimensional float array

    y : array-like of shape (n_samples, n_timesteps) or (n_samples, n_dims, n_timesteps)
        The collection of samples

    threshold : float
        The maximum threshold to consider a match

    dim : int, optional
        The time series dimension to search

    sample : int or array-like, optional
        The samples to compare to

        - if ``sample=None`` the distances to all samples in data is returned
        - if sample is an int the distance to that sample is returned
        - if sample is an array-like the distance to all samples in sample are returned
        - if ``n_samples=1``, ``samples`` is an int or ``len(samples)==1`` a scalar is returned
        - otherwise an array is returned

    metric : {'euclidean', 'scaled_euclidean'}, optional
        The distance metric

    metric_params: dict, optional
        Parameters to the metric

    return_distance : bool, optional
        - if `true` return the distance of the best match.

    Returns
    -------
    dist : list
        The distances of the matching positions

    matches : list
        The start position of the matches in each time series

    Warnings
    --------
    'scaled_dtw' is not supported.
    """
    return _distance.matches(
        x, y, threshold, dim, sample, metric, metric_params, return_distance
    )



import numpy as np
import pandas as pd
from sklearn.model_selection import cross_validate

from wildboar.datasets import load_dataset, list_datasets
from wildboar.ensemble import ShapeletForestClassifier, ExtraShapeletTreesClassifier
from wildboar.distance.dtw import dtw_distance
from sklearn.neighbors import KNeighborsClassifier

random_state = 1234
classifiers = {
    "Nearest neighbors": KNeighborsClassifier(
        n_neighbors=1,
        metric="euclidean",
    ),
    "Shapelet forest": ShapeletForestClassifier(
        n_shapelets=10,
        metric="scaled_euclidean",
        random_state=random_state,
        n_jobs=-1,
    ),
    "Extra shapelet trees": ExtraShapeletTreesClassifier(
        metric="scaled_euclidean",
        n_jobs=-1,
        random_state=random_state,
    ),
}

repository = "wildboar/ucr-tiny"
datasets = list_datasets(repository)
df = pd.DataFrame(columns=classifiers.keys(), index=datasets, dtype=np.float)
for dataset in datasets:
    print(dataset)
    x, y = load_dataset(dataset, repository=repository)
    for clf_name, clf in classifiers.items():
        print(" ", clf_name)
        score = cross_validate(clf, x, y, scoring="roc_auc_ovo", n_jobs=1)
        df.loc[dataset, clf_name] = np.mean(score["test_score"])

df.to_csv("../tab/classification_cmp.csv", float_format="%.3f")



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

import hashlib
import io
import os
import re
import sys
import zipfile
from abc import abstractmethod, ABCMeta

import numpy as np
import requests
from wildboar import __version__ as wildboar_version
from pkg_resources import parse_version
from scipy.io.arff import loadarff

DEFAULT_TAG = "default"


def _replace_placeholders(url, **kwargs):
    """Replace placeholder values of the format {key} with kwargs[key]

    Parameters
    ----------
    url : str
        The input string

    **kwargs : dict
        The key and values to replace

    Returns
    -------
    str : url with placeholder values replaced
    """
    for arg, value in kwargs.items():
        url = url.replace("{%s}" % arg, value)
    return url


def _check_integrity(bundle_file, hash_file):
    """Check the integrity of the downloaded or cached file

    Parameters
    ----------
    bundle_file : str, bytes or PathLike
        Path to the bundle file

    hash_file : str, bytes or PathLike
        Path to the hash file

    Returns
    -------
    bool : true if the hash of bundle file matches the contents of the hash file
    """
    with open(hash_file, "r") as f:
        hash = f.readline().strip()

    if hash is not None:
        actual_hash = _sha1(bundle_file)
        if hash != actual_hash:
            raise ValueError(
                "integrity check failed, expected '%s', got '%s'" % (hash, actual_hash)
            )
    return True


def _sha1_is_sane(hash_file):
    """Check the sanity of a hash file

    Parameters
    ----------
    hash_file : str, bytes or PathLike
        The path to the hash file

    Returns
    -------
    bool : Returns true if hash is 40 characters long; otherwise false.
    """
    with open(hash_file, "r") as f:
        return len(f.readline().strip()) == 40


def _load_archive(
    bundle_name,
    download_url,
    cache_dir,
    *,
    create_cache_dir=True,
    progress=True,
    force=False,
):
    """Load or download a bundle

    Parameters
    ----------
    bundle_name : str
        The name of the cached file

    download_url : str
        The download url to the bundle and hash file

    cache_dir : str
        The cache directory

    create_cache_dir : bool
        Create the cache directory if missing

    progress : bool
        Show progress information

    force : bool
        Remove any cached files and force re-download

    Returns
    -------
    archive : zipfile.ZipFile
        A zip-archive with datasets
    """
    if not os.path.exists(cache_dir):
        if create_cache_dir:
            os.makedirs(os.path.abspath(cache_dir), exist_ok=True)
        else:
            raise ValueError(
                "output directory does not exist (set create_cache_dir=True to create it)"
            )

    cached_hash = os.path.join(cache_dir, "%s.sha1" % bundle_name)
    cached_bundle = os.path.join(cache_dir, "%s.zip" % bundle_name)

    if force:
        if os.path.exists(cached_hash):
            os.remove(cached_hash)
        if os.path.exists(cached_bundle):
            os.remove(cached_bundle)

    if not os.path.exists(cached_hash) or not _sha1_is_sane(cached_hash):
        _download_hash_file(cached_hash, "%s.sha1" % download_url, bundle_name)

    if os.path.exists(cached_bundle) and _check_integrity(cached_bundle, cached_hash):
        try:
            return zipfile.ZipFile(open(cached_bundle, "rb"))
        except zipfile.BadZipFile:
            os.remove(cached_bundle)

    _download_bundle_file(cached_bundle, "%s.zip" % download_url, bundle_name, progress)
    _check_integrity(cached_bundle, cached_hash)
    return zipfile.ZipFile(open(cached_bundle, "rb"))


def _download_hash_file(cached_hash, hash_url, filename):
    """Download the

    Parameters
    ----------
    cached_hash : str, bytes or PathLike
        The path to the cached hash

    hash_url : str
        The download url

    filename : str
        The filename of the bundle
    """
    with open(cached_hash, "w") as f:
        response = requests.get(hash_url)
        if not response:
            f.close()
            os.remove(cached_hash)
            raise ValueError(
                "bundle (%s) not found (.sha1-file is missing). Try another version or tag."
                % filename
            )
        f.write(response.text)


def _download_bundle_file(cached_bundle, bundle_url, filename, progress):
    """Download the bundle

    Parameters
    ----------
    cached_bundle : str, bytes or PathLike
        The path to the cached bundle

    bundle_url : str
        The download url

    filename : str
        The filename of the bundle

    progress : bool
        Show progress bar
    """
    with open(cached_bundle, "wb") as f:
        response = requests.get(bundle_url, stream=True)
        if not response:
            f.close()
            os.remove(cached_bundle)
            raise ValueError(
                "bundle (%s) not found (.zip-file is missing). Try another version or tag."
                % filename
            )

        total_length = response.headers.get("content-length")
        if total_length is None:  # no content length header
            f.write(response.content)
        else:
            length = 0
            total_length = int(total_length)
            for data in response.iter_content(chunk_size=4096):
                length += len(data)
                f.write(data)
                done = int(50 * length / total_length)
                if length % 10 == 0 and progress:
                    sys.stderr.write(
                        "\r[%s%s] %d/%d downloading %s"
                        % (
                            "=" * done,
                            " " * (50 - done),
                            length,
                            total_length,
                            filename,
                        )
                    )
                    sys.stderr.flush()


class Repository(metaclass=ABCMeta):
    """A repository is a collection of bundles"""

    @property
    @abstractmethod
    def name(self):
        """Name of the repository

        Returns
        -------
        str : the name of the repository
        """
        pass

    @property
    @abstractmethod
    def version(self):
        """The repository version

        Returns
        -------
        str : the version of the repository
        """
        pass

    @property
    @abstractmethod
    def download_url(self):
        """The url template for downloading bundles

        Returns
        -------
        str : the download url
        """
        pass

    @property
    @abstractmethod
    def wildboar_requires(self):
        """The minimum required wildboar version

        Returns
        -------
        str : the min version
        """
        pass

    @abstractmethod
    def get_bundles(self):
        """Get all bundles

        Returns
        -------
        dict : a dictionary of key and bundle
        """
        pass

    def get_bundle(self, key):
        """Get a bundle with the specified key

        Parameters
        ----------
        key : str
            Key of the bundle

        Returns
        -------
        bundle : Bundle, optional
            A bundle or None
        """
        bundle = self.get_bundles().get(key)
        if bundle is None:
            raise ValueError("bundle (%s) does not exist" % key)
        return bundle

    def load_dataset(
        self,
        bundle,
        dataset,
        *,
        cache_dir,
        version=None,
        tag=None,
        create_cache_dir=True,
        progress=True,
        dtype=None,
        force=False,
    ):
        bundle = self.get_bundle(bundle)
        version = version or bundle.version
        tag = tag or DEFAULT_TAG

        cache_dir = os.path.join(cache_dir, self.name)
        download_url = _replace_placeholders(
            self.download_url, bundle=bundle.key, version=version, tag=tag
        )

        with _load_archive(
            bundle_name=bundle.get_filename(version, tag),
            download_url=download_url,
            cache_dir=cache_dir,
            create_cache_dir=create_cache_dir,
            progress=progress,
            force=force,
        ) as archive:
            return bundle.load(dataset, archive, dtype=dtype)

    def list_datasets(
        self,
        bundle,
        *,
        cache_dir,
        version=None,
        tag=None,
        create_cache_dir=True,
        progress=True,
        force=False,
    ):
        bundle = self.get_bundle(bundle)
        version = version or bundle.version
        tag = tag or DEFAULT_TAG

        cache_dir = os.path.join(cache_dir, self.name)
        download_url = _replace_placeholders(
            self.download_url, bundle=bundle.key, version=version, tag=tag
        )
        with _load_archive(
            bundle_name=bundle.get_filename(version, tag),
            download_url=download_url,
            cache_dir=cache_dir,
            create_cache_dir=create_cache_dir,
            progress=progress,
            force=force,
        ) as archive:
            return bundle.list(archive)

    def clear_cache(self, cache_dir, keep_last_version=True):
        cache_dir = os.path.join(cache_dir, self.name)
        if not os.path.exists(cache_dir) or not os.path.isdir(cache_dir):
            return

        keep = []
        if keep_last_version:
            keep = [
                bundle.get_filename(tag=DEFAULT_TAG)
                for _, bundle in self.get_bundles().items()
            ]

        for filename in os.listdir(cache_dir):
            basename, ext = os.path.splitext(filename)
            full_path = os.path.join(cache_dir, filename)
            if (
                os.path.isfile(full_path)
                and ext in [".zip", ".sha1"]
                and not basename in keep
            ):
                os.remove(full_path)

    def refresh(self):
        """Refresh the repository"""
        pass


def _validate_url(url):
    if "{bundle}" in url and "{version}" in url and "{tag}" in url:
        return url
    else:
        raise ValueError(
            "repository url is invalid, got %s ({bundle}, {version} and {tag} are required)"
            % url
        )


def _validate_repository_name(str):
    if re.match("[a-zA-Z]+", str):
        return str
    else:
        raise ValueError("repository name (%s) is not valid" % str)


def _validate_bundle_key(str):
    if re.match("[a-zA-Z0-9\-]+", str):
        return str
    else:
        raise ValueError("bundle key (%s) is not valid" % str)


def _validate_version(str):
    if re.match("(^(?:\d+\.)?(?:\d+\.)?(?:\*|\d+)$)", str):
        return str
    else:
        raise ValueError("version (%s) is not valid" % str)


class JSONRepository(Repository):
    def __init__(self, url):
        self.repo_url = url
        self.__refresh()

    @property
    def wildboar_requires(self):
        return self._wildboar_requires

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def download_url(self):
        return self._bundle_url

    def get_bundles(self):
        return self._bundles

    def refresh(self):
        self.__refresh()

    def __refresh(self):
        json = requests.get(self.repo_url).json()
        self._wildboar_requires = json["wildboar_requires"]
        self._name = _validate_repository_name(json["name"])
        self._version = _validate_version(json["version"])
        if parse_version(self.wildboar_requires) > parse_version(wildboar_version):
            raise ValueError(
                "repository requires wildboar (>=%s), got %s",
                self.wildboar_requires,
                wildboar_version,
            )
        self._bundle_url = _validate_url(json["bundle_url"])
        bundles = {}
        for bundle_json in json["bundles"]:
            key = _validate_bundle_key(bundle_json["key"])
            version = _validate_version(bundle_json["version"])
            name = bundle_json["name"]
            class_index = bundle_json["class_index"]
            description = bundle_json["description"]
            bundles[bundle_json["key"]] = NpyBundle(
                key=key,
                version=version,
                name=name,
                description=description,
                class_index=class_index,
            )

        self._bundles = bundles


class RepositoryCollection:
    def __init__(self):
        self.repositories = []

    def __getitem__(self, item):
        return next(
            (repository for repository in self.repositories if repository.name == item),
            None,
        )

    def __contains__(self, item):
        return any(r for r in self.repositories if r.name == item)

    def __iter__(self):
        return iter(self.repositories)

    def append(self, item):
        if self[item]:
            raise ValueError("cannot overwrite repository, %s" % item.name)
        self.repositories.append(item)


class Bundle(metaclass=ABCMeta):
    """Base class for handling dataset bundles

    Attributes
    ----------

    name : str
        Human-readable name of the bundle

    description : str
        Description of the bundle

    class_index : int or array-like
        Index of the class label(s)
    """

    def __init__(
        self,
        *,
        key,
        version,
        name,
        description=None,
        class_index=-1,
    ):
        """Construct a bundle

        Parameters
        ----------
        key : str
            A unique key of the bundle

        version : str
            The version of the bundle

        name : str
            Human-readable name of the bundle

        description : str
            Description of the bundle

        class_index : int or array-like
            Index of the class label(s)
        """
        self.key = key
        self.version = version
        self.name = name
        self.description = description
        self.class_index = class_index

    def get_filename(self, version=None, tag=None, ext=None):
        filename = "%s-v%s" % (self.key, version or self.version)
        if tag:
            filename += ":%s" % tag
        if ext:
            filename += ext
        return filename

    def list(self, archive):
        """List all datasets in this bundle

        Parameters
        ----------
        archive : ZipFile
            The bundle file

        Returns
        -------
        dataset_names : list
            A sorted list of datasets in the bundle
        """
        names = []
        for f in archive.filelist:
            path, ext = os.path.splitext(f.filename)
            if self._is_dataset(path, ext):
                filename = os.path.basename(path)
                filename = re.sub("_(TRAIN|TEST)", "", filename)
                names.append(filename)

        return sorted(set(names))

    def load(
        self,
        name,
        archive,
        *,
        dtype=None,
    ):
        """Load a dataset from the bundle

        Parameters
        ----------
        name : str
            Name of the dataset

        archive : ZipFile
            The zip-file bundle

        dtype : object, optional
            Cast the data and label matrix to a specific type

        Returns
        -------
        x : ndarray
            Data samples

        y : ndarray
            Data labels

        n_training_samples : int
            Number of samples that are for training. The value is <= x.shape[0]
        """
        dtype = dtype or np.float64
        datasets = []
        for dataset in map(_Dataset, archive.filelist):
            if dataset.filename == name and self._is_dataset(dataset.path, dataset.ext):
                datasets.append(dataset)

        if not datasets:
            raise ValueError("no dataset found (%s)" % name)
        train_parts = [
            self._load_array(archive, dataset.file)
            for dataset in datasets
            if dataset.part == "train"
        ]
        test_parts = [
            self._load_array(archive, dataset.file)
            for dataset in datasets
            if dataset.part == "test"
        ]

        data = np.vstack(train_parts)
        n_train_samples = data.shape[0]
        if test_parts:
            test = np.vstack(test_parts)
            data = np.vstack([data, test])

        y = data[:, self.class_index].astype(dtype)
        x = np.delete(data, self.class_index, axis=1).astype(dtype)
        return x, y, n_train_samples

    @abstractmethod
    def _is_dataset(self, file_name, ext):
        """Overridden by subclasses

        Check if a path and extension is to be considered a dataset. The check should be simple and only consider
        the filename and or extension of the file. Validation of the file should be deferred to `_load_array`

        Parameters
        ----------
        file_name : str
            Name of the dataset file

        ext : str
            Extension of the dataset file

        Returns
        -------
        bool
            True if it is a dataset
        """
        pass

    @abstractmethod
    def _load_array(self, archive, file):
        """Overridden by subclasses

        Load the file inside the archive and convert to a numpy array

        Parameters
        ----------
        archive : ZipFile
            The zip-archive in which the file reside

        file : str
            Path to the file inside the zip-archive

        Returns
        -------
        ndarray
            The dataset converted to a ndarray
        """
        pass


class ArffBundle(Bundle):
    """bundle of .arff-files"""

    def __init__(
        self,
        *,
        key,
        version,
        name,
        description=None,
        class_index=-1,
        encoding="utf-8",
    ):
        super().__init__(
            key=key,
            version=version,
            name=name,
            description=description,
            class_index=class_index,
        )
        self.encoding = encoding

    def _is_dataset(self, file_name, ext):
        return ext == ".arff"

    def _load_array(self, archive, file):
        with io.TextIOWrapper(archive.open(file), encoding=self.encoding) as io_wrapper:
            arff, _metadata = loadarff(io_wrapper)
            arr = np.array(arff.tolist())
            return arr


class NpyBundle(Bundle):
    """bundle of numpy binary files"""

    def _is_dataset(self, file_name, ext):
        return ext == ".npy"

    def _load_array(self, archive, file):
        return np.load(archive.open(file))


class _Dataset:
    def __init__(self, zip_info):
        self.file = zip_info
        self.path, self.ext = os.path.splitext(zip_info.filename)
        self.filename = os.path.basename(self.path)
        if "_TRAIN" in self.filename:
            self.part = "train"
            self.filename = self.filename.replace("_TRAIN", "")
        elif "_TEST" in self.filename:
            self.part = "test"
            self.filename = self.filename.replace("_TEST", "")
        else:
            self.part = "train"


def _sha1(file, buf_size=65536):
    sha1 = hashlib.sha1()
    with open(file, "rb") as f:
        while True:
            data = f.read(buf_size)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()



import numpy as np
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA

from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import cross_validate
from sklearn.pipeline import make_pipeline

from wildboar.datasets import load_dataset
from wildboar.ensemble import ShapeletForestEmbedding

random_state = 1234
x, y = load_dataset("GunPoint")
print(x.shape)

pipe = make_pipeline(
    ShapeletForestEmbedding(
        n_shapelets=1,
        min_shapelet_size=0,
        max_shapelet_size=1,
        metric="scaled_euclidean",
        sparse_output=True,
        max_depth=5,
        random_state=random_state,
        n_jobs=-1,
    ),
    LogisticRegression(solver="newton-cg", random_state=random_state),
)
cv = cross_validate(pipe, x, y, cv=5, scoring="accuracy", n_jobs=1)
err = 1 - np.mean(cv["test_score"])
print(err)

pca = make_pipeline(
    ShapeletForestEmbedding(
        metric="scaled_euclidean",
        sparse_output=False,
        max_depth=5,
        random_state=random_state,
    ),
    PCA(n_components=2, random_state=random_state),
)
p = pca.fit_transform(x)
var = pca.steps[1][1].explained_variance_ratio_

labels, index = np.unique(y, return_inverse=True)
colors = plt.cm.rainbow(np.linspace(0, 1, len(labels)))
plt.scatter(p[:, 0], p[:, 1], color=colors[index, :])
plt.xlabel("Component 1 (%.2f variance explained)" % var[0])
plt.ylabel("Component 2 (%.2f variance explained)" % var[1])
plt.title("PCA with 2 components  (E_{out} = %.2f )" % err)
plt.show()



import numpy as np

from wildboar.ensemble import ShapeletForestClassifier

# Example 2: multivariate shapelet forest

# fmt:off
x = [
    # Example 1
    [
        [0, 0, 1, 10, 1],  # dimension 1
        [0, 0, 1, 10, 1],  # dimension 2
    ],

    # Example 2
    [
        [0, 1, 9, 1, 0],  # dimension 1
        [1, 9, 1, 0, 0],  # dimension 2
    ],

    # etc...
    [[0, 1, 9, 1, 0], [0, 1, 2, 3, 4]],
    [[1, 2, 3, 0, 0], [0, 0, 0, 1, 2]],
    [[0, 0, -1, 0, 1], [1, 2, 3, 0, 1]],
]
# fmt:on

# `x` is an array of shape `[5, 2, 5]`, i.e., 5 examples with 2
# dimensions each consisting of 5 time steps
x = np.array(x, dtype=np.float64)
n_samples, n_dimensions, n_timesteps = x.shape

# `y` is the output target
y = np.array([0, 0, 1, 1, 0])

random_state = np.random.RandomState(123)
order = np.arange(n_samples)
random_state.shuffle(order)

x = x[order, :, :]
y = y[order]

f = ShapeletForestClassifier(random_state=random_state, n_shapelets=1)
f.fit(x, y)

predict = 1
print("Predicting the class of example:")
print(x[predict, :].reshape(-1, n_dimensions, n_timesteps))
print(
    "The true class is:",
    y[predict],
    "and the predicted class is:",
    f.predict(x[predict, :].reshape(1, n_dimensions, -1))[0],
)



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten

import abc

from sklearn.base import BaseEstimator

__all__ = ["BaseCounterfactual"]


class BaseCounterfactual(BaseEstimator):
    """Base estimator for counterfactual explanations"""

    @abc.abstractmethod
    def fit(self, estimator):
        """Fit the counterfactual to a given estimator

        Parameters
        ----------
        estimator : object
            An estimator for which counterfactual explanations are produced

        Returns
        -------
        self
        """
        pass

    @abc.abstractmethod
    def transform(self, x, y):
        """Transform the i:th sample in x to a sample that would be labeled as the i:th label in y

        Parameters
        ----------
        x : array-like of shape (n_samples, n_timestep) or (n_samples, n_dimension, n_timestep)
            The samples to generate counterfactual explanations for

        y : array-like of shape (n_samples,)
            The desired label of the counterfactual sample

        Returns
        -------

        counterfactuals : ndarray of same shape as x
            The counterfactual for each sample. If success[i] == False, then
            the value of counterfactuals[i] is undefined.

        success : ndarray of shape (n_samples,)
             Boolean vector indicating successful transformations.
        """
        pass



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten
import numbers
from abc import abstractmethod

import numpy as np
from joblib import Parallel, delayed
from sklearn.base import OutlierMixin
from sklearn.ensemble import BaggingClassifier
from sklearn.ensemble import BaggingRegressor
from sklearn.ensemble._bagging import BaseBagging
from sklearn.metrics import roc_curve, precision_recall_curve
from sklearn.preprocessing import OneHotEncoder
from sklearn.utils import check_array
from sklearn.utils import check_random_state
from sklearn.utils.fixes import _joblib_parallel_args
from sklearn.utils.validation import check_is_fitted

from scipy import sparse

from ..model_selection.outlier import threshold_score
from ..tree import ExtraShapeletTreeClassifier
from ..tree import ExtraShapeletTreeRegressor
from ..tree import ShapeletTreeClassifier
from ..tree import ShapeletTreeRegressor


class ShapeletForestMixin:
    def apply(self, x):
        x = self._validate_x_predict(x, check_input=True)
        results = Parallel(
            n_jobs=self.n_jobs,
            verbose=self.verbose,
            **self._parallel_args(),
        )(delayed(tree.apply)(x, check_input=False) for tree in self.estimators_)

        return np.array(results).T

    def decision_function(self, x):
        x = self._validate_x_predict(x, check_input=True)
        indicators = Parallel(
            n_jobs=self.n_jobs,
            verbose=self.verbose,
            **_joblib_parallel_args(prefer="threads"),
        )(
            delayed(tree.decision_path)(x, check_input=False)
            for tree in self.estimators_
        )

        n_nodes = [0]
        n_nodes.extend([i.shape[1] for i in indicators])
        n_nodes_ptr = np.array(n_nodes).cumsum()

        return sparse.hstack(indicators).tocsr(), n_nodes_ptr

    def _validate_x_predict(self, x, check_input):
        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimensions X.ndim ({})".format(x.ndim))
        if self.n_dims > 1 and x.ndim != 3:
            raise ValueError("illegal input dimensions X.ndim != 3")
        if x.shape[-1] != self.n_timestep:
            raise ValueError(
                "illegal input shape ({} != {})".format(x.shape[-1], self.n_timestep)
            )
        if x.ndim > 2 and x.shape[1] != self.n_dims:
            raise ValueError(
                "illegal input shape ({} != {}".format(x.shape[1], self.n_dims)
            )
        if check_input:
            x = check_array(x, dtype=np.float64, allow_nd=True, order="C")
        if x.dtype != np.float64 or not x.flags.c_contiguous:
            x = np.ascontiguousarray(x, dtype=np.float64)
        x = x.reshape(x.shape[0], self.n_dims * self.n_timestep)
        return x


class BaseShapeletForestClassifier(ShapeletForestMixin, BaggingClassifier):
    """Base class for shapelet forest classifiers.

    Warnings
    --------
    This class should not be used directly. Use derived classes
    instead.
    """

    def __init__(
        self,
        base_estimator,
        *,
        estimator_params=tuple(),
        oob_score=False,
        n_estimators=100,
        max_depth=None,
        min_samples_split=2,
        n_shapelets=10,
        min_shapelet_size=0,
        max_shapelet_size=1,
        metric="euclidean",
        metric_params=None,
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            base_estimator=base_estimator,
            n_estimators=n_estimators,
            bootstrap=bootstrap,
            bootstrap_features=False,
            n_jobs=n_jobs,
            warm_start=warm_start,
            random_state=random_state,
            oob_score=oob_score,
        )
        self.estimator_params = estimator_params
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.metric = metric
        self.metric_params = metric_params

    def _parallel_args(self):
        return _joblib_parallel_args(prefer="threads")

    def predict(self, X, check_input=True):
        x = self._validate_x_predict(X, check_input)
        return super().predict(X)

    def predict_proba(self, x, check_input=True):
        x = self._validate_x_predict(x, check_input)
        return super().predict_proba(x)

    def predict_log_proba(self, x, check_input=True):
        x = self._validate_x_predict(x, check_input)
        return super().predict_log_proba(x)

    def fit(self, x, y, sample_weight=None, check_input=True):
        """Fit a random shapelet forest classifier"""
        if check_input:
            x = check_array(x, dtype=np.float64, allow_nd=True, order="C")
            y = check_array(y, ensure_2d=False)

        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimension")

        n_samples = x.shape[0]
        self.n_timestep = x.shape[-1]
        if x.ndim > 2:
            n_dims = x.shape[1]
        else:
            n_dims = 1

        self.n_dims = n_dims

        if x.dtype != np.float64 or not x.flags.c_contiguous:
            x = np.ascontiguousarray(x, dtype=np.float64)

        if not y.flags.c_contiguous:
            y = np.ascontiguousarray(y, dtype=np.intp)

        x = x.reshape(n_samples, n_dims * self.n_timestep)
        self.n_features_in_ = x.shape[1]
        super()._fit(x, y, self.max_samples, self.max_depth, sample_weight)
        return self

    def _make_estimator(self, append=True, random_state=None):
        estimator = super()._make_estimator(append, random_state)
        if self.n_dims > 1:
            estimator.force_dim = self.n_dims
        return estimator


class ShapeletForestClassifier(BaseShapeletForestClassifier):
    """An ensemble of random shapelet tree classifiers.

    Examples
    --------

    >>> from wildboar.ensemble import ShapeletForestClassifier
    >>> from wildboar.datasets import load_synthetic_control
    >>> x, y = load_synthetic_control()
    >>> f = ShapeletForestClassifier(n_estimators=100, metric='scaled_euclidean')
    >>> f.fit(x, y)
    >>> y_hat = f.predict(x)
    """

    def __init__(
        self,
        *,
        n_estimators=100,
        n_shapelets=10,
        max_depth=None,
        min_samples_split=2,
        min_shapelet_size=0,
        max_shapelet_size=1,
        metric="euclidean",
        metric_params=None,
        oob_score=False,
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        random_state=None,
    ):
        """Shapelet forest classifier.

        Parameters
        ----------
        n_estimators : int, optional
            The number of estimators

        n_shapelets : int, optional
            The number of shapelets to sample at each node

        bootstrap : bool, optional
            Use bootstrap sampling to fit the base estimators

        n_jobs : int, optional
            The number of processor cores used for fitting the ensemble

        min_shapelet_size : float, optional
            The minimum shapelet size to sample

        max_shapelet_size : float, optional
            The maximum shapelet size to sample

        min_samples_split : int, optional
            The minimum samples required to split the decision trees

        warm_start : bool, optional
            When set to True, reuse the solution of the previous call to fit
            and add more estimators to the ensemble, otherwise, just fit
            a whole new ensemble.

        metric : {'euclidean', 'scaled_euclidean', 'scaled_dtw'}, optional
            Set the metric used to compute the distance between shapelet and time series

        metric_params : dict, optional
            Parameters passed to the metric construction

        oob_score : bool, optional
            Compute out-of-bag estimates of the ensembles performance.

        random_state : int or RandomState, optional
            Controls the random resampling of the original dataset and the construction of
            the base estimators. Pass an int for reproducible output across multiple
            function calls.
        """
        super().__init__(
            base_estimator=ShapeletTreeClassifier(),
            estimator_params=(
                "max_depth",
                "n_shapelets",
                "min_samples_split",
                "min_shapelet_size",
                "max_shapelet_size",
                "metric",
                "metric_params",
            ),
            n_shapelets=n_shapelets,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            oob_score=oob_score,
            bootstrap=bootstrap,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
        )


class ExtraShapeletTreesClassifier(BaseShapeletForestClassifier):
    """An ensemble of extremely random shapelet trees for time series regression.

    Examples
    --------

    >>> from wildboar.ensemble import ExtraShapeletTreesClassifier
    >>> from wildboar.datasets import load_synthetic_control
    >>> x, y = load_synthetic_control()
    >>> f = ExtraShapeletTreesClassifier(n_estimators=100, metric='scaled_euclidean')
    >>> f.fit(x, y)
    >>> y_hat = f.predict(x)

    """

    def __init__(
        self,
        *,
        n_estimators=100,
        max_depth=None,
        min_samples_split=2,
        min_shapelet_size=0,
        max_shapelet_size=1,
        metric="euclidean",
        metric_params=None,
        oob_score=False,
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        random_state=None,
    ):
        """Construct a extra shapelet trees classifier.

        Parameters
        ----------
        n_estimators : int, optional
            The number of estimators

        bootstrap : bool, optional
            Use bootstrap sampling to fit the base estimators

        n_jobs : int, optional
            The number of processor cores used for fitting the ensemble

        min_shapelet_size : float, optional
            The minimum shapelet size to sample

        max_shapelet_size : float, optional
            The maximum shapelet size to sample

        min_samples_split : int, optional
            The minimum samples required to split the decision trees

        warm_start : bool, optional
            When set to True, reuse the solution of the previous call to fit
            and add more estimators to the ensemble, otherwise, just fit
            a whole new ensemble.

        metric : {'euclidean', 'scaled_euclidean', 'scaled_dtw'}, optional
            Set the metric used to compute the distance between shapelet and time series

        metric_params : dict, optional
            Parameters passed to the metric construction

        random_state : int or RandomState, optional
            Controls the random resampling of the original dataset and the construction of
            the base estimators. Pass an int for reproducible output across multiple
            function calls.
        """
        super().__init__(
            base_estimator=ExtraShapeletTreeClassifier(),
            estimator_params=(
                "max_depth",
                "min_samples_split",
                "min_shapelet_size",
                "max_shapelet_size",
                "metric",
                "metric_params",
            ),
            n_shapelets=1,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            oob_score=oob_score,
            bootstrap=bootstrap,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
        )


class BaseShapeletForestRegressor(ShapeletForestMixin, BaggingRegressor):
    """Base class for shapelet forest regressors.

    Warnings
    --------
    This class should not be used directly. Use derived classes
    instead.
    """

    @abstractmethod
    def __init__(
        self,
        base_estimator,
        *,
        estimator_params=tuple(),
        oob_score=False,
        n_estimators=100,
        max_depth=None,
        min_samples_split=2,
        n_shapelets=10,
        min_shapelet_size=0.0,
        max_shapelet_size=1.0,
        metric="euclidean",
        metric_params=None,
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            base_estimator=base_estimator,
            n_estimators=n_estimators,
            bootstrap=bootstrap,
            bootstrap_features=False,
            n_jobs=n_jobs,
            warm_start=warm_start,
            random_state=random_state,
            oob_score=oob_score,
        )
        self.estimator_params = estimator_params
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.n_shapelets = n_shapelets
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.metric = metric
        self.metric_params = metric_params

    def _parallel_args(self):
        return _joblib_parallel_args(prefer="threads")

    def predict(self, x, check_input=True):
        x = self._validate_x_predict(x, check_input)
        return super().predict(x)

    def fit(self, x, y, sample_weight=None, check_input=True):
        """Fit a random shapelet forest regressor"""
        if check_input:
            x = check_array(x, dtype=np.float64, allow_nd=True, order="C")
            y = check_array(y, dtype=np.float64, ensure_2d=False, order="C")

        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimension")

        n_samples = x.shape[0]
        self.n_timestep = x.shape[-1]
        if x.ndim > 2:
            n_dims = x.shape[1]
        else:
            n_dims = 1

        self.n_dims = n_dims

        if x.dtype != np.float64 or not x.flags.c_contiguous:
            x = np.ascontiguousarray(x, dtype=np.float64)

        if y.dtype != np.float64 or not y.flags.c_contiguous:
            y = np.ascontiguousarray(y, dtype=np.float64)

        x = x.reshape(n_samples, n_dims * self.n_timestep)
        self.n_features_in_ = x.shape[1]
        super()._fit(
            x, y, self.max_samples, self.max_depth, sample_weight=sample_weight
        )
        return self

    def _make_estimator(self, append=True, random_state=None):
        estimator = super()._make_estimator(append, random_state)
        if self.n_dims > 1:
            estimator.force_dim = self.n_dims
        return estimator


class ShapeletForestRegressor(BaseShapeletForestRegressor):
    """An ensemble of random shapelet regression trees.

    Examples
    --------

    >>> from wildboar.ensemble import ShapeletForestRegressor
    >>> from wildboar.datasets import load_synthetic_control
    >>> x, y = load_synthetic_control()
    >>> f = ShapeletForestRegressor(n_estimators=100, metric='scaled_euclidean')
    >>> f.fit(x, y)
    >>> y_hat = f.predict(x)
    """

    def __init__(
        self,
        *,
        n_estimators=100,
        n_shapelets=10,
        max_depth=None,
        min_samples_split=2,
        min_shapelet_size=0,
        max_shapelet_size=1,
        metric="euclidean",
        metric_params=None,
        oob_score=False,
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        random_state=None,
    ):
        """Shapelet forest regressor.

        Parameters
        ----------
        n_estimators : int, optional
            The number of estimators

        n_shapelets : int, optional
            The number of shapelets to sample at each node

        bootstrap : bool, optional
            Use bootstrap sampling to fit the base estimators

        n_jobs : int, optional
            The number of processor cores used for fitting the ensemble

        min_shapelet_size : float, optional
            The minimum shapelet size to sample

        max_shapelet_size : float, optional
            The maximum shapelet size to sample

        min_samples_split : int, optional
            The minimum samples required to split the decision trees

        warm_start : bool, optional
            When set to True, reuse the solution of the previous call to fit
            and add more estimators to the ensemble, otherwise, just fit
            a whole new ensemble.

        metric : {'euclidean', 'scaled_euclidean', 'scaled_dtw'}, optional
            Set the metric used to compute the distance between shapelet and time series

        metric_params : dict, optional
            Parameters passed to the metric construction

        oob_score : bool, optional
            Compute out-of-bag estimates of the ensembles performance.

        random_state : int or RandomState, optional
            Controls the random resampling of the original dataset and the construction of
            the base estimators. Pass an int for reproducible output across multiple
            function calls.
        """
        super().__init__(
            base_estimator=ShapeletTreeRegressor(),
            estimator_params=(
                "max_depth",
                "n_shapelets",
                "min_samples_split",
                "min_shapelet_size",
                "max_shapelet_size",
                "metric",
                "metric_params",
            ),
            n_shapelets=n_shapelets,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            oob_score=oob_score,
            bootstrap=bootstrap,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
        )


class ExtraShapeletTreesRegressor(BaseShapeletForestRegressor):
    """An ensemble of extremely random shapelet trees for time series regression.

    Examples
    --------

    >>> from wildboar.ensemble import ExtraShapeletTreesRegressor
    >>> from wildboar.datasets import load_synthetic_control
    >>> x, y = load_synthetic_control()
    >>> f = ExtraShapeletTreesRegressor(n_estimators=100, metric='scaled_euclidean')
    >>> f.fit(x, y)
    >>> y_hat = f.predict(x)

    """

    def __init__(
        self,
        *,
        n_estimators=100,
        max_depth=None,
        min_samples_split=2,
        min_shapelet_size=0,
        max_shapelet_size=1,
        metric="euclidean",
        metric_params=None,
        oob_score=False,
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        random_state=None,
    ):
        """Construct a extra shapelet trees regressor.

        Parameters
        ----------
        n_estimators : int, optional
            The number of estimators

        bootstrap : bool, optional
            Use bootstrap sampling to fit the base estimators

        n_jobs : int, optional
            The number of processor cores used for fitting the ensemble

        min_shapelet_size : float, optional
            The minimum shapelet size to sample

        max_shapelet_size : float, optional
            The maximum shapelet size to sample

        min_samples_split : int, optional
            The minimum samples required to split the decision trees

        warm_start : bool, optional
            When set to True, reuse the solution of the previous call to fit
            and add more estimators to the ensemble, otherwise, just fit
            a whole new ensemble.

        metric : {'euclidean', 'scaled_euclidean', 'scaled_dtw'}, optional
            Set the metric used to compute the distance between shapelet and time series

        metric_params : dict, optional
            Parameters passed to the metric construction

        random_state : int or RandomState, optional
            Controls the random resampling of the original dataset and the construction of
            the base estimators. Pass an int for reproducible output across multiple
            function calls.
        """
        super().__init__(
            base_estimator=ExtraShapeletTreeRegressor(),
            estimator_params=(
                "max_depth",
                "min_samples_split",
                "min_shapelet_size",
                "max_shapelet_size",
                "metric",
                "metric_params",
            ),
            n_shapelets=1,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            oob_score=oob_score,
            bootstrap=bootstrap,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
        )


class ShapeletForestEmbedding(BaseShapeletForestRegressor):
    """An ensemble of random shapelet trees

    An unsupervised transformation of a time series dataset
    to a high-dimensional sparse representation. A time series i
    indexed by the leaf that it falls into. This leads to a binary
    coding of a time series with as many ones as trees in the forest.

    The dimensionality of the resulting representation is
    ``<= n_estimators * 2^max_depth``
    """

    def __init__(
        self,
        n_estimators=100,
        *,
        n_shapelets=1,
        max_depth=5,
        min_samples_split=2,
        min_shapelet_size=0,
        max_shapelet_size=1,
        metric="euclidean",
        metric_params=None,
        bootstrap=True,
        warm_start=False,
        n_jobs=None,
        sparse_output=True,
        random_state=None,
    ):
        """Construct a shapelet forest embedding.

        Parameters
        ----------

        n_estimators : int, optional
            The number of estimators

        bootstrap : bool, optional
            Use bootstrap sampling to fit the base estimators

        n_jobs : int, optional
            The number of processor cores used for fitting the ensemble

        min_shapelet_size : float, optional
            The minimum shapelet size to sample

        max_shapelet_size : float, optional
            The maximum shapelet size to sample

        min_samples_split : int, optional
            The minimum samples required to split the decision trees

        warm_start : bool, optional
            When set to True, reuse the solution of the previous call to fit
            and add more estimators to the ensemble, otherwise, just fit
            a whole new ensemble.

        metric : {'euclidean', 'scaled_euclidean', 'scaled_dtw'}, optional
            Set the metric used to compute the distance between shapelet and time series

        metric_params : dict, optional
            Parameters passed to the metric construction

        sparse_output : bool, optional
            Return a sparse CSR-matrix.

        random_state : int or RandomState, optional
            Controls the random resampling of the original dataset and the construction of
            the base estimators. Pass an int for reproducible output across multiple
            function calls.
        """
        super().__init__(
            base_estimator=ExtraShapeletTreeRegressor(),
            estimator_params=(
                "n_shapelets",
                "max_depth",
                "min_samples_split",
                "min_shapelet_size",
                "max_shapelet_size",
                "metric",
                "metric_params",
            ),
            n_shapelets=n_shapelets,
            n_estimators=n_estimators,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_shapelet_size=min_shapelet_size,
            max_shapelet_size=max_shapelet_size,
            metric=metric,
            metric_params=metric_params,
            oob_score=False,
            bootstrap=bootstrap,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.sparse_output = sparse_output

    def _set_oob_score(self, X, y):
        raise NotImplementedError("OOB score not supported")

    def fit(self, x, y=None, sample_weight=None, check_input=True):
        self.fit_transform(x, y, sample_weight, check_input)
        return self

    def fit_transform(self, x, y=None, sample_weight=None, check_input=True):
        random_state = check_random_state(self.random_state)
        if check_input:
            x = check_array(x, dtype=np.float64, allow_nd=True, order="C")

        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimension")

        n_samples = x.shape[0]
        self.n_timestep = x.shape[-1]
        if x.ndim > 2:
            n_dims = x.shape[1]
        else:
            n_dims = 1

        self.n_dims = n_dims

        if x.dtype != np.float64 or not x.flags.c_contiguous:
            x = np.ascontiguousarray(x, dtype=np.float64)

        if n_dims > 1:
            self.base_estimator_.force_dim = n_dims

        x = x.reshape(n_samples, n_dims * self.n_timestep)
        y = random_state.uniform(size=x.shape[0])
        super().fit(x, y, sample_weight=sample_weight)
        self.one_hot_encoder_ = OneHotEncoder(
            sparse=self.sparse_output, handle_unknown="ignore"
        )
        return self.one_hot_encoder_.fit_transform(self.apply(x))

    def transform(self, x):
        check_is_fitted(self)
        return self.one_hot_encoder_.transform(self.apply(x))


class IsolationShapeletForest(ShapeletForestMixin, OutlierMixin, BaseBagging):
    """A isolation shapelet forest.

    .. versionadded:: 0.3.5

    Attributes
    ----------
    offset_ : float
        The offset for computing the final decision

    Examples
    --------

    >>> from wildboar.ensemble import IsolationShapeletForest
    >>> from wildboar.datasets import load_two_lead_ecg
    >>> from model_selection.outlier import train_test_split
    >>> from sklearn.metrics import balanced_accuracy_score
    >>> x, y = load_two_lead_ecg("two_lead_ecg")
    >>> x_train, x_test, y_train, y_test = train_test_split(x, y, 1, test_size=0.2, anomalies_train_size=0.05)
    >>> f = IsolationShapeletForest(n_estimators=100, contamination=balanced_accuracy_score)
    >>> f.fit(x_train, y_train)
    >>> y_pred = f.predict(x_test)
    >>> balanced_accuracy_score(y_test, y_pred)

    Or using default offset threshold

    >>> from wildboar.ensemble import IsolationShapeletForest
    >>> from wildboar.datasets import load_two_lead_ecg
    >>> from model_selection.outlier import train_test_split
    >>> from sklearn.metrics import balanced_accuracy_score
    >>> f = IsolationShapeletForest()
    >>> x, y = load_two_lead_ecg("two_lead_ecg")
    >>> x_train, x_test, y_train, y_test = train_test_split(x, y, 1, test_size=0.2, anomalies_train_size=0.05)
    >>> f.fit(x_train)
    >>> y_pred = f.predict(x_test)
    >>> balanced_accuracy_score(y_test, y_pred)
    """

    def __init__(
        self,
        *,
        n_estimators=100,
        bootstrap=False,
        n_jobs=None,
        min_shapelet_size=0,
        max_shapelet_size=1,
        min_samples_split=2,
        max_samples="auto",
        contamination="auto",
        contamination_set="training",
        warm_start=False,
        metric="euclidean",
        metric_params=None,
        random_state=None,
    ):
        """Construct a shapelet isolation forest

        Parameters
        ----------
        n_estimators : int, optional
            The number of estimators

        bootstrap : bool, optional
            Use bootstrap sampling to fit the base estimators

        n_jobs : int, optional
            The number of processor cores used for fitting the ensemble

        min_shapelet_size : float, optional
            The minimum shapelet size to sample

        max_shapelet_size : float, optional
            The maximum shapelet size to sample

        min_samples_split : int, optional
            The minimum samples required to split the decision trees

        max_samples : float or int
            The number of samples to draw to train each base estimator

        contamination : str, float or callable
            The strategy for computing the offset (see `offset_`)

            - if 'auto' ``offset_=-0.5``
            - if 'auc' ``offset_`` is computed as the offset that maximizes the
              area under ROC in the training or out-of-bag set (see ``contamination_set``).
            - if 'prc' ``offset_`` is computed as the offset that maximizes the
              area under PRC in the training or out-of-bag set (see ``contamination_set``)
            - if callable ``offset_`` is computed as the offset that maximizes the score
              computed by the callable in training or out-of-bag set (see ``contamination_set``)
            - if float ``offset_`` is computed as the c:th percentile of scores in the training
              or out-of-bag set (see ``contamination_set``)

            Setting contamination to either 'auc' or 'prc' require that `y` is passed to `fit`.

        contamination_set : {'training', 'oob'}, optional
            Compute the ``offset_`` from either the out-of-bag samples or the training samples.
            'oob' require `bootstrap=True`.

        warm_start : bool, optional
            When set to True, reuse the solution of the previous call to fit
            and add more estimators to the ensemble, otherwise, just fit
            a whole new ensemble.

        metric : {'euclidean', 'scaled_euclidean', 'scaled_dtw'}, optional
            Set the metric used to compute the distance between shapelet and time series

        metric_params : dict, optional
            Parameters passed to the metric construction

        random_state : int or RandomState, optional
            Controls the random resampling of the original dataset and the construction of
            the base estimators. Pass an int for reproducible output across multiple
            function calls.
        """
        super(IsolationShapeletForest, self).__init__(
            base_estimator=ExtraShapeletTreeRegressor(),
            bootstrap=bootstrap,
            bootstrap_features=False,
            warm_start=warm_start,
            n_jobs=n_jobs,
            n_estimators=n_estimators,
            random_state=random_state,
        )
        self.estimator_params = (
            "min_samples_split",
            "min_shapelet_size",
            "max_shapelet_size",
            "metric",
            "metric_params",
        )
        self.metric = metric
        self.metric_params = metric_params
        self.min_shapelet_size = min_shapelet_size
        self.max_shapelet_size = max_shapelet_size
        self.min_samples_split = min_samples_split
        self.contamination = contamination
        self.contamination_set = contamination_set
        self.max_samples = max_samples
        self.n_dims = None
        self.n_timestep = None

    def _set_oob_score(self, x, y):
        raise NotImplementedError("OOB score not supported")

    def _parallel_args(self):
        return _joblib_parallel_args(prefer="threads")

    def fit(self, x, y=None, sample_weight=None, check_input=True):
        random_state = check_random_state(self.random_state)
        if check_input:
            x = check_array(x, dtype=np.float64, allow_nd=True, order="C")

        if x.ndim < 2 or x.ndim > 3:
            raise ValueError("illegal input dimension")

        n_samples = x.shape[0]
        self.n_timestep = x.shape[-1]
        if x.ndim > 2:
            n_dims = x.shape[1]
        else:
            n_dims = 1

        self.n_dims = n_dims

        if x.dtype != np.float64 or not x.flags.c_contiguous:
            x = np.ascontiguousarray(x, dtype=np.float64)

        if n_dims > 1:
            self.base_estimator_.force_dim = n_dims

        x = x.reshape(n_samples, n_dims * self.n_timestep)
        rnd_y = random_state.uniform(size=x.shape[0])
        if isinstance(self.max_samples, str):
            if self.max_samples == "auto":
                max_samples = min(x.shape[0], 256)
            else:
                raise ValueError(
                    "max_samples (%s) is not supported." % self.max_samples
                )
        elif isinstance(self.max_samples, numbers.Integral):
            max_samples = min(self.max_samples, x.shape[0])
        else:
            if not 0.0 < self.max_samples <= 1.0:
                raise ValueError(
                    "max_samples must be in (0, 1], got %r" % self.max_samples
                )
            max_samples = int(self.max_samples * x.shape[0])

        max_depth = int(np.ceil(np.log2(max(max_samples, 2))))
        self.n_features_in_ = self.n_timestep
        super(IsolationShapeletForest, self)._fit(
            x,
            rnd_y,
            max_samples=max_samples,
            max_depth=max_depth,
            sample_weight=sample_weight,
        )

        self.max_samples_ = max_samples
        if self.contamination == "auto":
            self.offset_ = -0.5
        elif self.contamination in ["auc", "prc"] or hasattr(
            self.contamination, "__call__"
        ):
            if y is None:
                raise ValueError(
                    "contamination cannot be computed without training labels"
                )

            if self.contamination_set == "oob":
                if not self.bootstrap:
                    raise ValueError(
                        "contamination cannot be computed from oob-samples unless bootstrap=True"
                    )
                scores = self._oob_score_samples(x)
            else:
                scores = self.score_samples(x)

            if self.contamination == "auc":
                fpr, tpr, thresholds = roc_curve(y, scores)
                best_threshold = np.argmax(tpr - fpr)
            elif self.contamination == "prc":
                precision, recall, thresholds = precision_recall_curve(y, scores)
                fscore = (2 * precision * recall) / (precision + recall)
                best_threshold = np.argmax(fscore)
            else:
                score = threshold_score(y, scores, self.contamination)
                best_threshold = np.argmax(score)
                thresholds = scores
            self.offset_ = thresholds[best_threshold]
        elif isinstance(self.contamination, numbers.Real):
            if not 0.0 < self.contamination <= 1.0:
                raise ValueError(
                    "contamination must be in (0, 1], got %r" % self.contamination
                )
            if self.contamination_set == "training":
                self.offset_ = np.percentile(
                    self.score_samples(x), 100.0 * self.contamination
                )
            elif self.contamination_set == "oob":
                if not self.bootstrap:
                    raise ValueError(
                        "contamination cannot be computed from oob-samples unless bootstrap=True"
                    )
                self.offset_ = np.percentile(
                    self._oob_score_samples(x), 100.0 * self.contamination
                )
            else:
                raise ValueError(
                    "contamination_set (%s) is not supported" % self.contamination_set
                )
        else:
            raise ValueError(
                "contamination (%s) is not supported." % self.contamination
            )

        return self

    def predict(self, x):
        is_inlier = np.ones(x.shape[0])
        is_inlier[self.decision_function(x) < 0] = -1
        return is_inlier

    def decision_function(self, x):
        return self.score_samples(x) - self.offset_

    def score_samples(self, x):
        check_is_fitted(self)
        x = self._validate_x_predict(x, check_input=True)
        return _score_samples(x, self.estimators_, self.max_samples_)

    def _oob_score_samples(self, x):
        n_samples = x.shape[0]
        score_samples = np.zeros((n_samples,))

        for i in range(x.shape[0]):
            estimators = []
            for estimator, samples in zip(self.estimators_, self.estimators_samples_):
                if i not in samples:
                    estimators.append(estimator)
            score_samples[i] = _score_samples(
                x[i].reshape((1, self.n_dims, self.n_timestep)),
                estimators,
                self.max_samples_,
            )
        return score_samples


def _score_samples(x, estimators, max_samples):
    # From: https://github.com/scikit-learn/scikit-learn/blob/0fb307bf39bbdacd6ed713c00724f8f871d60370/sklearn/ensemble/_iforest.py#L411
    depths = np.zeros(x.shape[0], order="f")

    for tree in estimators:
        leaves_index = tree.apply(x)
        node_indicator = tree.decision_path(x)
        n_samples_leaf = tree.tree_.n_node_samples[leaves_index]

        depths += (
            np.ravel(node_indicator.sum(axis=1))
            + _average_path_length(n_samples_leaf)
            - 1.0
        )
    scores = 2 ** (
        -depths / (len(estimators) * _average_path_length(np.array([max_samples])))
    )
    return -scores


def _average_path_length(n_samples_leaf):
    # From: https://github.com/scikit-learn/scikit-learn/blob/0fb307bf39bbdacd6ed713c00724f8f871d60370/sklearn/ensemble/_iforest.py#L480
    n_samples_leaf_shape = n_samples_leaf.shape
    n_samples_leaf = n_samples_leaf.reshape((1, -1))
    average_path_length = np.zeros(n_samples_leaf.shape)

    mask_1 = n_samples_leaf <= 1
    mask_2 = n_samples_leaf == 2
    not_mask = ~np.logical_or(mask_1, mask_2)

    average_path_length[mask_1] = 0.0
    average_path_length[mask_2] = 1.0
    average_path_length[not_mask] = (
        2.0 * (np.log(n_samples_leaf[not_mask] - 1.0) + np.euler_gamma)
        - 2.0 * (n_samples_leaf[not_mask] - 1.0) / n_samples_leaf[not_mask]
    )

    return average_path_length.reshape(n_samples_leaf_shape)



# This file is part of wildboar
#
# wildboar is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wildboar is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Isak Samsten
import warnings

import numpy as np

from sklearn.neighbors import KNeighborsClassifier
from sklearn.utils.validation import check_is_fitted
from sklearn.metrics.pairwise import paired_distances

from wildboar.ensemble import ShapeletForestClassifier, ExtraShapeletTreesClassifier

from ._nn import KNeighborsCounterfactual
from ._proto import PrototypeCounterfactual
from ._sf import ShapeletForestCounterfactual

__all__ = [
    "counterfactuals",
    "score",
    "ShapeletForestCounterfactual",
    "KNeighborsCounterfactual",
    "PrototypeCounterfactual",
]

_COUNTERFACTUAL_EXPLAINER = {
    "prototype": PrototypeCounterfactual,
}


def _infer_counterfactual(estimator):
    """Infer the counterfactual explainer to use based on the estimator

    Parameters
    ----------
    estimator : object
        The estimator

    Returns
    -------
    BaseCounterfactual
        The counterfactual transformer
    """
    if isinstance(estimator, (ShapeletForestClassifier, ExtraShapeletTreesClassifier)):
        return ShapeletForestCounterfactual
    elif isinstance(estimator, KNeighborsClassifier):
        return KNeighborsCounterfactual
    else:
        return PrototypeCounterfactual


def score(x_true, x_counterfactuals, metric="euclidean", success=None):
    """Compute the score for the counterfactuals

    Parameters
    ----------
    x_true : array-like of shape (n_samples, n_timestep)
        The true samples

    x_counterfactuals : array-like of shape (n_samples, n_timestep)
        The counterfactual samples

    metric : str, callable, list or dict, optional
        The scoring metric

        - if str use metrics from scikit-learn
        - if list compute all metrics and return a dict where the key is
          the name of the metric and the value an ndarray of scores
        - if dict compute all metrics and return a dict where the key is
          the key and the value an ndarry of scores
        - if callable

    success : ndarray of shape (n_samples)
        Indicator matrix of successful counterfactual transformations

    Returns
    -------
    score : ndarray or dict
        The scores
    """
    if success:
        x_true = x_true[success]
        x_counterfactuals = x_counterfactuals[success]

    if isinstance(metric, str) or hasattr(metric, "__call__"):
        return paired_distances(x_true, x_counterfactuals, metric=metric)
    else:
        sc = {}
        if isinstance(metric, dict):
            for key, value in metric.items():
                sc[key] = paired_distances(x_true, x_counterfactuals, metric=value)
        elif isinstance(metric, list):
            for item in metric:
                sc[item] = paired_distances(x_true, x_counterfactuals, metric=item)
        else:
            raise ValueError("invalid metric, got %r" % metric)
        return sc


def counterfactuals(
    estimator,
    x,
    y,
    *,
    method="infer",
    scoring=None,
    valid_scoring=False,
    random_state=None,
    **kwargs
):
    """Compute a single counterfactual example for each sample

    Parameters
    ----------
    estimator : object
        The estimator used to compute the counterfactual example

    x : array-like of shape (n_samples, n_timestep) or (n_samples, n_dimension, n_timestep)
        The data samples to fit counterfactuals to

    y : array-like broadcast to shape (n_samples,)
        The desired label of the counterfactual

    method : str, optional
        The method to generate counterfactual explanations

        - if 'infer', infer the most appropriate counterfactual explanation method
          based on the estimator
        - if 'prototype', compute model agnostic counterfactual explanations using
          the PrototypeCounterfactual method

    scoring : str, callable, list or dict, optional
        The scoring function to determine the goodness of

    valid_scoring : bool, optional
        Only compute score for successful counterfactuals

    random_state : RandomState or int, optional
        The pseudo random number generator to ensure stable result

    **kwargs : dict, optional
        Optional arguments to the counterfactual explainer

    Returns
    -------
    x_counterfactuals : ndarray of shape (n_samples, n_timestep) or (n_samples, n_dimension, n_timestep)
        The counterfactual example.

    valid : ndarray of shape (n_samples,)
        Indicator matrix for valid counterfactuals

    score : ndarray of shape (n_samples,) or dict, optional
        Score of the counterfactual transform. Only returned if ``scoring`` is not None
    """
    check_is_fitted(estimator)
    if method == "infer":
        Explainer = _infer_counterfactual(estimator)
        if Explainer == PrototypeCounterfactual:
            warnings.warn(
                "no specific counterfactual explanation method is available for the given estimator."
            )
            if not ("background_x" in kwargs or "background_y" in kwargs):
                raise ValueError("background_x and background_y are required")
    else:
        Explainer = _COUNTERFACTUAL_EXPLAINER[method]

    if Explainer is None:
        raise ValueError("no counterfactual explainer for '%r'" % method)

    y = np.broadcast_to(y, (x.shape[0],))
    explainer = Explainer(**kwargs)
    explainer.set_params(random_state=random_state)
    explainer.fit(estimator)
    x_counterfactuals, success = explainer.transform(x, y)
    if scoring:
        sc = score(
            x,
            x_counterfactuals,
            metric=scoring,
            success=success if valid_scoring else None,
        )
        return x_counterfactuals, success, sc
    else:
        return x_counterfactuals, success



import numpy as np
from wildboar.datasets import (
    load_dataset,
    list_datasets,
    list_repositories,
    list_bundles,
    load_datasets,
)

print(list_datasets(repository="wildboar/ucr:no-missing"))
print(list_datasets(repository="wildboar/outlier:1.0:hard"))
print(list_repositories())
print(list_bundles("wildboar"))

for dataset, (x, y) in load_datasets(
    repository="wildboar/ucr-tiny",
    filter=["n_timestep>10", "n_samples<=200", "n_labels<=3"],
):
    print(dataset)

load_dataset("GunPoint", repository="wildboar/ucr-tiny")
